# coding=utf-8

from functools import wraps


class AsyncDecoratorMixin(object):
    """Provide the ability to decorate both sync and async functions."""

    is_async = False

    @classmethod
    def async_(cls, operation, tags=None, **kwargs):
        """
        Instrument an async function via a decorator.

        This will return an awaitable which must be awaited.
        Using this on a synchronous function will raise a
        RuntimeError.

        ``
           @instrument.async_("Foo")
           async def foo():
              ...
        ``
        """
        instance = cls(operation, tags=tags, **kwargs)
        instance.is_async = True
        return instance

    def __call__(self, func):
        if self.is_async:
            # Until https://bugs.python.org/issue37398 has a resolution,
            # manually wrap the async function
            @wraps(func)
            async def decorated(*args, **kwds):
                with self._recreate_cm():
                    return await func(*args, **kwds)

            return decorated
        else:
            return super().__call__(func)



# coding=utf-8

import logging

import scout_apm.core
from scout_apm.core.error import ErrorMonitor
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import filter_element

try:
    from fastmcp.server.middleware import Middleware, MiddlewareContext
except ImportError:
    Middleware = None
    MiddlewareContext = None

logger = logging.getLogger(__name__)


class ScoutMiddleware(Middleware if Middleware is not None else object):
    """
    FastMCP middleware for Scout APM instrumentation.

    This middleware tracks tool executions and collects performance metrics
    for FastMCP servers.
    """

    def __init__(self):
        if Middleware is not None:
            super().__init__()
        installed = scout_apm.core.install()
        self._do_nothing = not installed

    async def on_call_tool(self, context, call_next):
        """
        Track tool execution with Scout APM.

        This hook is called whenever a tool is executed. We create a
        TrackedRequest for each tool invocation to measure performance
        and capture errors.
        """
        if self._do_nothing:
            return await call_next(context)

        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True

        # Get tool name from execution context
        tool_name = getattr(context.message, "name", "unknown")
        operation = f"Controller/{tool_name}"
        tracked_request.operation = operation

        # Add rich metadata from tool object via context
        try:
            tool = await context.fastmcp_context.fastmcp.get_tool(tool_name)
            self._tag_tool_metadata(tracked_request, tool)
        except Exception as exc:
            # Tool not found or other error - continue without metadata
            logger.warning(f"Unable to fetch tool metadata for {tool_name}: {exc}")

        # Tag tool arguments (filtered for sensitive data)
        arguments = getattr(context.message, "arguments", {})
        if arguments:
            filtered_args = filter_element("", arguments)
            tracked_request.tag("arguments", str(filtered_args))

        with tracked_request.span(operation=operation, should_capture_backtrace=False):
            try:
                result = await call_next(context)
                return result
            except Exception as exc:
                tracked_request.tag("error", "true")
                ErrorMonitor.send(
                    (type(exc), exc, exc.__traceback__),
                    custom_controller=operation,
                    custom_params={"tool": tool_name, "arguments": arguments},
                )
                raise

    def _tag_tool_metadata(self, tracked_request, tool):
        """
        Add rich tool metadata as tags to the tracked request.

        This extracts metadata from the FastMCP tool object and adds it
        as tags for filtering and analysis in Scout APM.
        https://gofastmcp.com/servers/tools#decorator-arguments
        """
        # Add tool description (truncated)
        if hasattr(tool, "description") and tool.description:
            tracked_request.tag("tool_description", str(tool.description)[:200])

        if hasattr(tool, "tags") and tool.tags:
            tracked_request.tag("tool_tags", ",".join(sorted(tool.tags)))

        # Add behavioral annotations (dict and object-style)
        # https://gofastmcp.com/servers/tools#param-annotations
        if hasattr(tool, "annotations") and tool.annotations:
            annotations = tool.annotations
            if isinstance(annotations, dict):
                tracked_request.tag("read_only", annotations.get("readOnlyHint", False))
                tracked_request.tag(
                    "destructive", annotations.get("destructiveHint", False)
                )
                tracked_request.tag(
                    "idempotent", annotations.get("idempotentHint", False)
                )
                tracked_request.tag("external", annotations.get("openWorldHint", True))
            else:
                # Object-style access
                if hasattr(annotations, "readOnlyHint"):
                    tracked_request.tag("read_only", annotations.readOnlyHint)
                if hasattr(annotations, "destructiveHint"):
                    tracked_request.tag("destructive", annotations.destructiveHint)
                if hasattr(annotations, "idempotentHint"):
                    tracked_request.tag("idempotent", annotations.idempotentHint)
                if hasattr(annotations, "openWorldHint"):
                    tracked_request.tag("external", annotations.openWorldHint)

        if hasattr(tool, "meta") and tool.meta:
            tracked_request.tag("tool_meta", str(tool.meta))



# coding=utf-8

import sys

import django
from django.apps import AppConfig
from django.conf import settings
from django.core.signals import got_request_exception
from django.test.signals import setting_changed

if django.VERSION < (3, 1):
    from django.views.debug import get_safe_settings
else:
    from django.views.debug import SafeExceptionReporterFilter

    def get_safe_settings():
        return SafeExceptionReporterFilter().get_safe_settings()


import scout_apm.core
from scout_apm.core.config import scout_config
from scout_apm.core.error import ErrorMonitor
from scout_apm.django.instruments.huey import ensure_huey_instrumented
from scout_apm.django.instruments.sql import ensure_sql_instrumented
from scout_apm.django.instruments.template import ensure_templates_instrumented
from scout_apm.django.request import get_request_components


class ScoutApmDjangoConfig(AppConfig):
    name = "scout_apm"
    verbose_name = "Scout Apm (Django)"

    def ready(self):
        self.update_scout_config_from_django_settings()
        setting_changed.connect(self.on_setting_changed)

        # Finish installing the agent. If the agent isn't installed for any
        # reason, return without installing instruments
        installed = scout_apm.core.install()
        if not installed:
            return

        if scout_config.value("errors_enabled"):
            got_request_exception.connect(self.on_got_request_exception)

        self.install_middleware()

        # Setup Instruments
        ensure_huey_instrumented()
        ensure_sql_instrumented()
        ensure_templates_instrumented()

    def update_scout_config_from_django_settings(self, **kwargs):
        for name in dir(settings):
            self.on_setting_changed(name)

    def on_got_request_exception(self, request, **kwargs):
        """
        Process this exception with the error monitoring solution.
        """
        ErrorMonitor.send(
            sys.exc_info(),
            request_components=get_request_components(request),
            request_path=request.path,
            request_params=dict(request.GET.lists()),
            session=dict(request.session.items())
            if hasattr(request, "session")
            else None,
            environment=get_safe_settings(),
        )

    def on_setting_changed(self, setting, **kwargs):
        cast = None
        if setting == "BASE_DIR":
            scout_name = "application_root"
            cast = str
        elif setting.startswith("SCOUT_"):
            scout_name = setting.replace("SCOUT_", "").lower()
        else:
            return

        try:
            value = getattr(settings, setting)
        except AttributeError:
            # It was removed
            scout_config.unset(scout_name)
        else:
            if cast is not None:
                value = cast(value)
            scout_config.set(**{scout_name: value})

    def install_middleware(self):
        """
        Attempts to insert the ScoutApm middleware as the first middleware
        (first on incoming requests, last on outgoing responses).
        """
        from django.conf import settings

        # If MIDDLEWARE is set, update that, with handling of tuple vs array forms
        if getattr(settings, "MIDDLEWARE", None) is not None:
            timing_middleware = "scout_apm.django.middleware.MiddlewareTimingMiddleware"
            view_middleware = "scout_apm.django.middleware.ViewTimingMiddleware"

            if isinstance(settings.MIDDLEWARE, tuple):
                if timing_middleware not in settings.MIDDLEWARE:
                    settings.MIDDLEWARE = (timing_middleware,) + settings.MIDDLEWARE
                if view_middleware not in settings.MIDDLEWARE:
                    settings.MIDDLEWARE = settings.MIDDLEWARE + (view_middleware,)
            else:
                if timing_middleware not in settings.MIDDLEWARE:
                    settings.MIDDLEWARE.insert(0, timing_middleware)
                if view_middleware not in settings.MIDDLEWARE:
                    settings.MIDDLEWARE.append(view_middleware)

        # Otherwise, we're doing old style middleware, do the same thing with
        # the same handling of tuple vs array forms
        else:
            timing_middleware = (
                "scout_apm.django.middleware.OldStyleMiddlewareTimingMiddleware"
            )
            view_middleware = "scout_apm.django.middleware.OldStyleViewMiddleware"

            if isinstance(settings.MIDDLEWARE_CLASSES, tuple):
                if timing_middleware not in settings.MIDDLEWARE_CLASSES:
                    settings.MIDDLEWARE_CLASSES = (
                        timing_middleware,
                    ) + settings.MIDDLEWARE_CLASSES

                if view_middleware not in settings.MIDDLEWARE_CLASSES:
                    settings.MIDDLEWARE_CLASSES = settings.MIDDLEWARE_CLASSES + (
                        view_middleware,
                    )
            else:
                if timing_middleware not in settings.MIDDLEWARE_CLASSES:
                    settings.MIDDLEWARE_CLASSES.insert(0, timing_middleware)
                if view_middleware not in settings.MIDDLEWARE_CLASSES:
                    settings.MIDDLEWARE_CLASSES.append(view_middleware)



# coding=utf-8

import logging
import warnings

import falcon

from scout_apm.api import install
from scout_apm.core.config import scout_config
from scout_apm.core.queue_time import track_request_queue_time
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import create_filtered_path, ignore_path

logger = logging.getLogger(__name__)

# Falcon Middleware docs:
# https://falcon.readthedocs.io/en/stable/api/middleware.html


class ScoutMiddleware(object):
    """
    Falcon Middleware for integration with Scout APM.
    """

    def __init__(self, config, hug_http_interface=None):
        self.api = None
        self.hug_http_interface = hug_http_interface
        installed = install(config=config)
        self._do_nothing = not installed

    def set_api(self, api):
        if not isinstance(api, falcon.API):
            raise ValueError("api should be an instance of falcon.API")
        self.api = api

    def process_request(self, req, resp):
        if self._do_nothing:
            return
        if self.api is None and self.hug_http_interface is not None:
            self.api = self.hug_http_interface.falcon
        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        req.context.scout_tracked_request = tracked_request
        tracked_request.start_span(
            operation="Middleware", should_capture_backtrace=False
        )

        path = req.path
        # Falcon URL parameter values are *either* single items or lists
        url_params = [
            (k, v)
            for k, vs in req.params.items()
            for v in (vs if isinstance(vs, list) else [vs])
        ]
        tracked_request.tag("path", create_filtered_path(path, url_params))
        if ignore_path(path):
            tracked_request.tag("ignore_transaction", True)

        if scout_config.value("collect_remote_ip"):
            # Determine a remote IP to associate with the request. The value is
            # spoofable by the requester so this is not suitable to use in any
            # security sensitive context.
            user_ip = (
                req.get_header("x-forwarded-for", default="").split(",")[0]
                or req.get_header("client-ip", default="").split(",")[0]
                or req.remote_addr
            )
            tracked_request.tag("user_ip", user_ip)

        queue_time = req.get_header("x-queue-start", default="") or req.get_header(
            "x-request-start", default=""
        )
        track_request_queue_time(queue_time, tracked_request)

    def process_resource(self, req, resp, resource, params):
        if self._do_nothing:
            return

        tracked_request = getattr(req.context, "scout_tracked_request", None)
        if tracked_request is None:
            # Somehow we didn't start a request - this might occur in
            # combination with a pretty adversarial application, so guard
            # against it, although if a request was started and the context was
            # lost, other problems might occur.
            return

        if self.api is None:
            warnings.warn(
                (
                    "{}.set_api() should be called before requests begin for"
                    + " more detail."
                ).format(self.__class__.__name__),
                RuntimeWarning,
                stacklevel=2,
            )
            operation = "Controller/{}.{}.{}".format(
                resource.__module__, resource.__class__.__name__, req.method
            )
        else:
            # Find the current responder's name. Falcon passes middleware the
            # current resource but unfortunately not the method being called, hence
            # we have to go through routing again.
            responder, _params, _resource, _uri_template = self.api._get_responder(req)
            operation = self._name_operation(req, responder, resource)

        span = tracked_request.start_span(
            operation=operation, should_capture_backtrace=False
        )
        tracked_request.operation = operation
        req.context.scout_resource_span = span

    def _name_operation(self, req, responder, resource):
        try:
            last_part = responder.__name__
        except AttributeError:
            last_part = req.method
        return "Controller/{}.{}.{}".format(
            resource.__module__, resource.__class__.__name__, last_part
        )

    def process_response(self, req, resp, resource, req_succeeded):
        tracked_request = getattr(req.context, "scout_tracked_request", None)
        if tracked_request is None:
            # Somehow we didn't start a request
            return

        # Falcon only stores the response status line, we have to parse it
        try:
            status_code = int(resp.status.split(" ")[0])
        except ValueError:
            # Bad status line - force it to be tagged as an error because
            # client will experience it as one
            status_code = 500

        if not req_succeeded or 500 <= status_code <= 599:
            tracked_request.tag("error", "true")

        span = getattr(req.context, "scout_resource_span", None)
        if span is not None:
            tracked_request.stop_span()
        # Stop Middleware span
        tracked_request.stop_span()



# coding=utf-8

import logging

import psutil

logger = logging.getLogger(__name__)


def get_rss_in_mb():
    rss_in_bytes = psutil.Process().memory_info().rss
    return rss_in_bytes / (1024 * 1024)


class Memory(object):
    metric_type = "Memory"
    metric_name = "Physical"
    human_name = "Process Memory"

    def run(self):
        value = get_rss_in_mb()
        logger.debug("%s: #%s", self.human_name, value)
        return value



# coding=utf-8

from django.conf import settings
from django.urls import get_urlconf

from scout_apm.core.config import scout_config
from scout_apm.core.queue_time import track_request_queue_time
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import create_filtered_path, ignore_path
from scout_apm.django.request import get_controller_name


def track_request_view_data(request, tracked_request):
    path = request.path
    tracked_request.tag(
        "path",
        create_filtered_path(
            path, [(k, v) for k, vs in request.GET.lists() for v in vs]
        ),
    )
    if ignore_path(path):
        tracked_request.tag("ignore_transaction", True)

    if scout_config.value("collect_remote_ip"):
        try:
            # Determine a remote IP to associate with the request. The value is
            # spoofable by the requester so this is not suitable to use in any
            # security sensitive context.
            user_ip = (
                request.META.get("HTTP_X_FORWARDED_FOR", "").split(",")[0]
                or request.META.get("HTTP_CLIENT_IP", "").split(",")[0]
                or request.META.get("REMOTE_ADDR", None)
            )
            tracked_request.tag("user_ip", user_ip)
        except Exception:
            pass

    # Django's request.user caches in this attribute on first access. We only
    # want to track the user if the application code has touched request.user
    # because touching it causes session access, which adds "Cookie" to the
    # "Vary" header.
    user = getattr(request, "_cached_user", None)
    if user is not None:
        try:
            tracked_request.tag("username", user.get_username())
        except Exception:
            pass

    tracked_request.tag("urlconf", get_urlconf(settings.ROOT_URLCONF))


class MiddlewareTimingMiddleware(object):
    """
    Insert as early into the Middleware stack as possible (outermost layers),
    so that other middlewares called after can be timed.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if not scout_config.value("monitor"):
            return self.get_response(request)

        tracked_request = TrackedRequest.instance()

        queue_time = request.META.get("HTTP_X_QUEUE_START") or request.META.get(
            "HTTP_X_REQUEST_START", ""
        )
        track_request_queue_time(queue_time, tracked_request)

        with tracked_request.span(
            operation="Middleware",
            should_capture_backtrace=False,
        ):
            return self.get_response(request)


class ViewTimingMiddleware(object):
    """
    Insert as deep into the middleware stack as possible, ideally wrapping no
    other middleware. Designed to time the View itself
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        """
        Wrap a single incoming request with start and stop calls.
        This will start timing, but relies on the process_view callback to
        capture more details about what view was really called, and other
        similar info.

        If process_view isn't called, then the request will not
        be recorded.  This can happen if a middleware further along the stack
        doesn't call onward, and instead returns a response directly.
        """
        if not scout_config.value("monitor"):
            return self.get_response(request)

        tracked_request = TrackedRequest.instance()

        # This operation name won't be recorded unless changed later in
        # process_view
        with tracked_request.span(operation="Unknown", should_capture_backtrace=False):
            response = self.get_response(request)
            track_request_view_data(request, tracked_request)
            if 500 <= response.status_code <= 599:
                tracked_request.tag("error", "true")
            return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        """
        Capture details about the view_func that is about to execute
        """
        if not scout_config.value("monitor"):
            return
        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True

        span = tracked_request.current_span()
        if span is not None:
            span.operation = get_controller_name(request)
            tracked_request.operation = span.operation

    def process_exception(self, request, exception):
        """
        Mark this request as having errored out

        Does not modify or catch or otherwise change the exception thrown
        """
        if not scout_config.value("monitor"):
            return
        TrackedRequest.instance().tag("error", "true")


class OldStyleMiddlewareTimingMiddleware(object):
    """
    Insert as early into the Middleware stack as possible (outermost layers),
    so that other middlewares called after can be timed.
    """

    def process_request(self, request):
        if not scout_config.value("monitor"):
            return
        tracked_request = TrackedRequest.instance()
        request._scout_tracked_request = tracked_request

        queue_time = request.META.get("HTTP_X_QUEUE_START") or request.META.get(
            "HTTP_X_REQUEST_START", ""
        )
        track_request_queue_time(queue_time, tracked_request)

        tracked_request.start_span(
            operation="Middleware", should_capture_backtrace=False
        )

    def process_response(self, request, response):
        # Only stop span if there's a request, but presume we are balanced,
        # i.e. that custom instrumentation within the application is not
        # causing errors
        tracked_request = getattr(request, "_scout_tracked_request", None)
        if tracked_request is not None:
            if 500 <= response.status_code <= 599:
                tracked_request.tag("error", "true")
            tracked_request.stop_span()
        return response


class OldStyleViewMiddleware(object):
    def process_view(self, request, view_func, view_func_args, view_func_kwargs):
        tracked_request = getattr(request, "_scout_tracked_request", None)
        if tracked_request is None:
            # Looks like OldStyleMiddlewareTimingMiddleware didn't run, so
            # don't do anything
            return

        tracked_request.is_real_request = True

        span = tracked_request.start_span(
            operation=get_controller_name(request), should_capture_backtrace=False
        )
        # Save the span into the request, so we can check
        # if we're matched up when stopping
        request._scout_view_span = span

    def process_response(self, request, response):
        tracked_request = getattr(request, "_scout_tracked_request", None)
        if tracked_request is None:
            # Looks like OldStyleMiddlewareTimingMiddleware didn't run, so
            # don't do anything
            return response

        track_request_view_data(request, tracked_request)

        # Only stop span if we started, but presume we are balanced, i.e. that
        # custom instrumentation within the application is not causing errors
        span = getattr(request, "_scout_view_span", None)
        if span is not None:
            tracked_request.stop_span()
        return response

    def process_exception(self, request, exception):
        tracked_request = getattr(request, "_scout_tracked_request", None)
        if tracked_request is None:
            # Looks like OldStyleMiddlewareTimingMiddleware didn't run, so
            # don't do anything
            return

        tracked_request.tag("error", "true")



# coding=utf-8

import logging

import wrapt

from scout_apm.core.tracked_request import TrackedRequest

try:
    import redis
except ImportError:  # pragma: no cover
    redis = None
else:
    if redis.VERSION[0] >= 3:
        from redis import Redis
        from redis.client import Pipeline
    else:  # pragma: no cover
        from redis import StrictRedis as Redis
        from redis.client import BasePipeline as Pipeline

logger = logging.getLogger(__name__)


have_patched_redis_execute_command = False
have_patched_pipeline_execute = False


def ensure_installed():
    global have_patched_redis_execute_command, have_patched_pipeline_execute

    logger.debug("Instrumenting redis.")

    if redis is None:
        logger.debug("Couldn't import redis - probably not installed.")
    else:
        if not have_patched_redis_execute_command:
            try:
                Redis.execute_command = wrapped_execute_command(Redis.execute_command)
            except Exception as exc:
                logger.warning(
                    "Failed to instrument redis.Redis.execute_command: %r",
                    exc,
                    exc_info=exc,
                )
            else:
                have_patched_redis_execute_command = True

        if not have_patched_pipeline_execute:
            try:
                Pipeline.execute = wrapped_execute(Pipeline.execute)
            except Exception as exc:
                logger.warning(
                    "Failed to instrument redis.Pipeline.execute: %r", exc, exc_info=exc
                )
            else:
                have_patched_pipeline_execute = True

    return True


@wrapt.decorator
def wrapped_execute_command(wrapped, instance, args, kwargs):
    try:
        op = args[0]
    except (IndexError, TypeError):
        op = "Unknown"

    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation="Redis/{}".format(op)):
        return wrapped(*args, **kwargs)


@wrapt.decorator
def wrapped_execute(wrapped, instance, args, kwargs):
    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation="Redis/MULTI"):
        return wrapped(*args, **kwargs)



# coding=utf-8

import django

if django.VERSION < (3, 2, 0):
    # Only define default_app_config when using a version earlier than 3.2
    default_app_config = "scout_apm.django.apps.ScoutApmDjangoConfig"



# coding=utf-8

import threading
import time
from threading import local as ThreadLocal

from scout_apm.core.tracked_request import TrackedRequest

try:
    from asgiref.local import Local as AsgiRefLocal
except ImportError:
    # Old versions of Python or asgiref < 3.1
    AsgiRefLocal = None

try:
    import asyncio
except ImportError:
    asyncio = None

try:
    from contextvars import ContextVar

    scout_context_var = ContextVar("__scout_trackedrequest")
except ImportError:
    scout_context_var = None


SCOUT_REQUEST_ATTR = "__scout_trackedrequest"


def get_current_asyncio_task():
    """
    Cross-version implementation of asyncio.current_task()
    Returns None if there is no task.
    """
    if asyncio:
        try:
            if hasattr(asyncio, "current_task"):
                # Python 3.7 and up
                return asyncio.current_task()
            else:
                # Python 3.6
                return asyncio.Task.current_task()
        except RuntimeError:
            return None


class SimplifiedAsgirefLocal:
    """
    A copy of asgiref 3.1+'s Local class without the sync_to_async /
    async_to_sync compatibility.
    """

    CLEANUP_INTERVAL = 60  # seconds

    def __init__(self):
        self._storage = {}
        self._last_cleanup = time.time()
        self._clean_lock = threading.Lock()

    def _get_context_id(self):
        """
        Get the ID we should use for looking up variables
        """
        # First, pull the current task if we can
        context_id = get_current_asyncio_task()
        # OK, let's try for a thread ID
        if context_id is None:
            context_id = threading.current_thread()
        return context_id

    def _cleanup(self):
        """
        Cleans up any references to dead threads or tasks
        """
        for key in list(self._storage.keys()):
            if isinstance(key, threading.Thread):
                if not key.is_alive():
                    del self._storage[key]
            elif isinstance(key, asyncio.Task):
                if key.done():
                    del self._storage[key]
        self._last_cleanup = time.time()

    def _maybe_cleanup(self):
        """
        Cleans up if enough time has passed
        """
        if time.time() - self._last_cleanup > self.CLEANUP_INTERVAL:
            with self._clean_lock:
                self._cleanup()

    def __getattr__(self, key):
        context_id = self._get_context_id()
        if key in self._storage.get(context_id, {}):
            return self._storage[context_id][key]
        else:
            raise AttributeError("%r object has no attribute %r" % (self, key))

    def __setattr__(self, key, value):
        if key in ("_storage", "_last_cleanup", "_clean_lock", "_thread_critical"):
            return super().__setattr__(key, value)
        self._maybe_cleanup()
        self._storage.setdefault(self._get_context_id(), {})[key] = value

    def __delattr__(self, key):
        context_id = self._get_context_id()
        if key in self._storage.get(context_id, {}):
            del self._storage[context_id][key]
        else:
            raise AttributeError("%r object has no attribute %r" % (self, key))


class LocalContext(object):
    def __init__(self):
        if AsgiRefLocal is not None:
            self._local = AsgiRefLocal()
        elif asyncio is not None:
            self._local = SimplifiedAsgirefLocal()
        else:
            self._local = ThreadLocal()
        self.use_context_var = scout_context_var is not None

    def get_tracked_request(self):
        if scout_context_var:
            if not scout_context_var.get(None):
                scout_context_var.set(TrackedRequest())
            return scout_context_var.get()
        if not hasattr(self._local, "tracked_request"):
            self._local.tracked_request = TrackedRequest()
        return self._local.tracked_request

    def clear_tracked_request(self, instance):
        if getattr(self._local, "tracked_request", None) is instance:
            del self._local.tracked_request
        if scout_context_var and scout_context_var.get(None) is instance:
            scout_context_var.set(None)


context = LocalContext()



# coding=utf-8

import json
import logging
import os
import socket
import struct
import threading
import time

from scout_apm.compat import queue
from scout_apm.core.agent.commands import Register
from scout_apm.core.agent.manager import get_socket_path
from scout_apm.core.config import scout_config
from scout_apm.core.threading import SingletonThread

# Time unit - monkey-patched in tests to make them run faster
SECOND = 1

logger = logging.getLogger(__name__)


class CoreAgentSocketThread(SingletonThread):
    _instance_lock = threading.Lock()
    _stop_event = threading.Event()
    _command_queue = queue.Queue(maxsize=500)

    @classmethod
    def _on_stop(cls):
        super(CoreAgentSocketThread, cls)._on_stop()
        # Unblock _command_queue.get()
        try:
            cls._command_queue.put(None, False)
        except queue.Full:
            pass

    @classmethod
    def send(cls, command):
        try:
            cls._command_queue.put(command, False)
        except queue.Full as exc:
            # TODO mark the command as not queued?
            logger.debug("CoreAgentSocketThread error on send: %r", exc, exc_info=exc)

        cls.ensure_started()

    @classmethod
    def wait_until_drained(cls, timeout_seconds=2.0, callback=None):
        interval_seconds = min(timeout_seconds, 0.05)
        start = time.time()
        while True:
            queue_size = cls._command_queue.qsize()
            queue_empty = queue_size == 0
            elapsed = time.time() - start
            if queue_empty or elapsed >= timeout_seconds:
                break

            if callback is not None:
                callback(queue_size)
                callback = None

            cls.ensure_started()

            time.sleep(interval_seconds)
        return queue_empty

    def run(self):
        self.socket_path = get_socket_path()
        self.socket = self.make_socket()

        try:
            self._connect()
            self._register()
            while True:
                try:
                    body = self._command_queue.get(block=True, timeout=1 * SECOND)
                except queue.Empty:
                    body = None

                if body is not None:
                    result = self._send(body)
                    if result:
                        self._command_queue.task_done()
                    else:
                        # Something was wrong with the socket.
                        self._disconnect()
                        self._connect()
                        self._register()

                # Check for stop event after each read. This allows opening,
                # sending, and then immediately stopping. We do this for
                # the metadata event at application start time.
                if self._stop_event.is_set():
                    logger.debug("CoreAgentSocketThread stopping.")
                    break
        except Exception as exc:
            logger.debug("CoreAgentSocketThread exception: %r", exc, exc_info=exc)
        finally:
            self.socket.close()
            logger.debug("CoreAgentSocketThread stopped.")

    def _send(self, command):
        msg = command.message()

        try:
            data = json.dumps(msg)
        except (ValueError, TypeError) as exc:
            logger.debug(
                "Exception when serializing command message: %r", exc, exc_info=exc
            )
            return False

        full_data = struct.pack(">I", len(data)) + data.encode("utf-8")
        try:
            self.socket.sendall(full_data)
        except OSError as exc:
            logger.debug(
                (
                    "CoreAgentSocketThread exception on _send:"
                    + " %r on PID: %s on thread: %s"
                ),
                exc,
                os.getpid(),
                threading.current_thread(),
                exc_info=exc,
            )
            return False

        # TODO do something with the response sent back in reply to command
        self._read_response()

        return True

    def _read_response(self):
        try:
            raw_size = self.socket.recv(4)
            if len(raw_size) != 4:
                # Ignore invalid responses
                return None
            size = struct.unpack(">I", raw_size)[0]
            message = bytearray(0)

            while len(message) < size:
                recv = self.socket.recv(size)
                message += recv

            return message
        except OSError as exc:
            logger.debug(
                "CoreAgentSocketThread error on read response: %r", exc, exc_info=exc
            )
            return None

    def _register(self):
        self._send(
            Register(
                app=scout_config.value("name"),
                key=scout_config.value("key"),
                hostname=scout_config.value("hostname"),
            )
        )

    def _connect(self, connect_attempts=5, retry_wait_secs=1):
        for attempt in range(1, connect_attempts + 1):
            logger.debug(
                (
                    "CoreAgentSocketThread attempt %d, connecting to %s, "
                    + "PID: %s, Thread: %s"
                ),
                attempt,
                self.socket_path,
                os.getpid(),
                threading.current_thread(),
            )
            try:
                self.socket.connect(self.get_socket_address())
                self.socket.settimeout(3 * SECOND)
                logger.debug("CoreAgentSocketThread connected")
                return
            except socket.error as exc:
                logger.debug(
                    "CoreAgentSocketThread connection error: %r", exc, exc_info=exc
                )
                # Return without waiting when reaching the maximum number of attempts.
                if attempt == connect_attempts:
                    raise
                time.sleep(retry_wait_secs * SECOND)

    def _disconnect(self):
        logger.debug("CoreAgentSocketThread disconnecting from %s", self.socket_path)
        try:
            self.socket.close()
        except socket.error as exc:
            logger.debug(
                "CoreAgentSocketThread exception on disconnect: %r", exc, exc_info=exc
            )
        finally:
            self.socket = self.make_socket()

    def make_socket(self):
        if self.socket_path.is_tcp:
            family = socket.AF_INET
        else:
            family = socket.AF_UNIX
        return socket.socket(family, socket.SOCK_STREAM)

    def get_socket_address(self):
        if self.socket_path.is_tcp:
            host, _, port = self.socket_path.tcp_address.partition(":")
            return host, int(port)
        return self.socket_path



# coding=utf-8

import errno
import hashlib
import json
import logging
import os
import subprocess
import tarfile
import time

from urllib3.exceptions import HTTPError

from scout_apm.compat import urllib3_cert_pool_manager
from scout_apm.core.config import scout_config

logger = logging.getLogger(__name__)

CA_ALREADY_RUNNING_EXIT_CODE = 3


class CoreAgentManager(object):
    def __init__(self):
        self.core_agent_bin_path = None
        self.core_agent_bin_version = None
        self.core_agent_dir = "{}/{}".format(
            scout_config.value("core_agent_dir"),
            scout_config.value("core_agent_full_name"),
        )
        self.downloader = CoreAgentDownloader(
            self.core_agent_dir, scout_config.value("core_agent_full_name")
        )

    def launch(self):
        if not scout_config.value("core_agent_launch"):
            logger.debug(
                "Not attempting to launch Core Agent "
                "due to 'core_agent_launch' setting."
            )
            return False

        if not self.verify():
            if not scout_config.value("core_agent_download"):
                logger.debug(
                    "Not attempting to download Core Agent due "
                    "to 'core_agent_download' setting."
                )
                return False

            self.download()

            if not self.verify():
                logger.debug("Failed to verify Core Agent. Not launching Core Agent.")
                return False

        return self.run()

    def download(self):
        self.downloader.download()

    def run(self):
        try:
            with open(os.devnull) as devnull:
                subprocess.check_call(
                    (
                        self.agent_binary()
                        + self.daemonize_flag()
                        + self.log_level()
                        + self.log_file()
                        + self.config_file()
                        + self.socket_path()
                    ),
                    close_fds=True,
                    stdout=devnull,
                )
        except subprocess.CalledProcessError as err:
            if err.returncode == CA_ALREADY_RUNNING_EXIT_CODE:
                # Other processes may have already started the core agent.
                logger.debug("Core agent already running.")
                return True
            else:
                logger.exception("CalledProcessError running Core Agent")
            return False
        except Exception:
            # TODO detect failure of launch properly
            logger.exception("Error running Core Agent")
            return False
        return True

    def agent_binary(self):
        return [self.core_agent_bin_path, "start"]

    def daemonize_flag(self):
        return ["--daemonize", "true"]

    def socket_path(self):
        path = get_socket_path()
        if path.is_tcp:
            return ["--tcp", path.tcp_address]
        else:
            return ["--socket", path]

    def log_level(self):
        # Old deprecated name "log_level"
        log_level = scout_config.value("log_level")
        if log_level is None:
            log_level = scout_config.value("core_agent_log_level")
        return ["--log-level", log_level]

    def log_file(self):
        # Old deprecated name "log_file"
        path = scout_config.value("log_file")
        if path is None:
            path = scout_config.value("core_agent_log_file")

        if path is not None:
            return ["--log-file", path]
        else:
            return []

    def config_file(self):
        # Old deprecated name "config_file"
        path = scout_config.value("config_file")
        if path is None:
            path = scout_config.value("core_agent_config_file")

        if path is not None:
            return ["--config-file", path]
        else:
            return []

    def verify(self):
        manifest = parse_manifest(self.core_agent_dir + "/manifest.json")
        if manifest is None:
            logger.debug(
                "Core Agent verification failed: CoreAgentManifest is not valid."
            )
            self.core_agent_bin_path = None
            self.core_agent_bin_version = None
            return False

        bin_path = os.path.join(self.core_agent_dir, manifest.bin_name)
        if sha256_digest(bin_path) == manifest.sha256:
            self.core_agent_bin_path = bin_path
            self.core_agent_bin_version = manifest.bin_version
            return True
        else:
            logger.debug("Core Agent verification failed: SHA mismatch.")
            self.core_agent_bin_path = None
            self.core_agent_bin_version = None
            return False


class CoreAgentDownloader(object):
    def __init__(self, download_destination, core_agent_full_name):
        self.stale_download_secs = 120
        self.destination = download_destination
        self.core_agent_full_name = core_agent_full_name
        self.package_location = self.destination + "/{}.tgz".format(
            self.core_agent_full_name
        )
        self.download_lock_path = self.destination + "/download.lock"
        self.download_lock_fd = None

    def download(self):
        self.create_core_agent_dir()
        self.obtain_download_lock()
        if self.download_lock_fd is not None:
            try:
                downloaded = self.download_package()
                if downloaded:
                    self.untar()
                else:
                    logger.debug("Core Agent download failed: bad http response.")
            except (OSError, HTTPError):
                logger.exception("Exception raised while downloading Core Agent")
            finally:
                self.release_download_lock()

    def create_core_agent_dir(self):
        try:
            os.makedirs(self.destination, scout_config.core_agent_permissions())
        except OSError:
            pass

    def obtain_download_lock(self):
        self.clean_stale_download_lock()
        try:
            self.download_lock_fd = os.open(
                self.download_lock_path,
                os.O_RDWR | os.O_CREAT | os.O_EXCL | os.O_NONBLOCK,
            )
        except OSError as exc:
            logger.debug(
                "Could not obtain download lock on %s",
                self.download_lock_path,
                exc_info=exc,
            )
            self.download_lock_fd = None

    def clean_stale_download_lock(self):
        try:
            delta = time.time() - os.stat(self.download_lock_path).st_ctime
            if delta > self.stale_download_secs:
                logger.debug("Clearing stale download lock file.")
                os.unlink(self.download_lock_path)
        except OSError:
            pass

    def release_download_lock(self):
        if self.download_lock_fd is not None:
            os.unlink(self.download_lock_path)
            os.close(self.download_lock_fd)

    def download_package(self):
        full_url = self.full_url()
        logger.debug("Downloading: %s to %s", full_url, self.package_location)
        http = urllib3_cert_pool_manager()
        response = http.request(
            "GET", full_url, preload_content=False, timeout=10.0, retries=3
        )
        try:
            if response.status != 200:
                return False
            with open(self.package_location, "wb") as fp:
                for chunk in response.stream():
                    fp.write(chunk)
        finally:
            response.release_conn()
            http.clear()
        return True

    def untar(self):
        try:
            t = tarfile.open(self.package_location, "r")
            # https://docs.python.org/3.12/library/tarfile.html#extraction-filters
            t.extractall(path=self.destination, filter="data")
        except Exception as e:
            logger.error(
                f"Error extracting {self.package_location} to {self.destination}: {e}"
            )

    def full_url(self):
        return "{root_url}/{core_agent_full_name}.tgz".format(
            root_url=self.root_url(), core_agent_full_name=self.core_agent_full_name
        )

    def root_url(self):
        return scout_config.value("download_url")


def parse_manifest(path):
    try:
        manifest_file = open(path)
    except OSError as exc:
        if exc.errno == errno.ENOENT:
            logger.debug("Core Agent Manifest does not exist at %s", path)
        else:
            logger.debug("Error opening Core Agent Manifest at %s", path, exc_info=exc)
        return None

    try:
        with manifest_file:
            data = json.load(manifest_file)
            logger.debug("Core Agent manifest json: %s", data)

            bin_name = data["core_agent_binary"]
            if not isinstance(bin_name, str):
                raise TypeError("core_agent_binary should be a string.")
            bin_version = data["core_agent_version"]
            if not isinstance(bin_version, str):
                raise TypeError("core_agent_version should be a string.")
            sha256 = data["core_agent_binary_sha256"]
            if not isinstance(sha256, str):
                raise TypeError("core_agent_binary_sha256 should be a string.")

            return CoreAgentManifest(
                bin_name=bin_name,
                bin_version=bin_version,
                sha256=sha256,
            )

    # IOError => OSError on Python 3
    except (KeyError, ValueError, TypeError, OSError, IOError) as exc:  # noqa: B014
        logger.debug("Error parsing Core Agent Manifest", exc_info=exc)
        return None


class CoreAgentManifest(object):
    __slots__ = ("bin_name", "bin_version", "sha256")

    def __init__(self, bin_name, bin_version, sha256):
        self.bin_name = bin_name
        self.bin_version = bin_version
        self.sha256 = sha256


def sha256_digest(filename, block_size=65536):
    try:
        sha256 = hashlib.sha256()
        with open(filename, "rb") as f:
            for block in iter(lambda: f.read(block_size), b""):
                sha256.update(block)
        return sha256.hexdigest()
    except OSError as exc:
        logger.debug("Error on digest", exc_info=exc)
        return None


class SocketPath(str):
    @property
    def is_tcp(self):
        return self.startswith("tcp://")

    @property
    def tcp_address(self):
        return self[len("tcp://") :]


def get_socket_path():
    # Old deprecated name "socket_path"
    socket_path = scout_config.value("socket_path")
    if socket_path is None:
        socket_path = scout_config.value("core_agent_socket_path")
    return SocketPath(socket_path)



# coding=utf-8

import logging
import sys

import wrapt
from flask import current_app
from flask.globals import request, session

import scout_apm.core
from scout_apm.core.config import scout_config
from scout_apm.core.error import ErrorMonitor
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import RequestComponents, werkzeug_track_request_data

logger = logging.getLogger(__name__)


class ScoutApm(object):
    def __init__(self, app):
        self.app = app
        self._attempted_install = False
        app.full_dispatch_request = self.wrapped_full_dispatch_request(
            app.full_dispatch_request
        )
        app.preprocess_request = self.wrapped_preprocess_request(app.preprocess_request)

    @wrapt.decorator
    def wrapped_full_dispatch_request(self, wrapped, instance, args, kwargs):
        if not self._attempted_install:
            self.extract_flask_settings()
            installed = scout_apm.core.install()
            self._do_nothing = not installed
            self._attempted_install = True

        if self._do_nothing:
            return wrapped(*args, **kwargs)

        # Pass on routing exceptions (normally 404's)
        if request.routing_exception is not None:
            return wrapped(*args, **kwargs)

        request_components = get_request_components(self.app, request)
        operation = "Controller/{}.{}".format(
            request_components.module, request_components.controller
        )

        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        tracked_request.operation = operation
        request._scout_tracked_request = tracked_request

        werkzeug_track_request_data(request, tracked_request)

        with tracked_request.span(
            operation=operation, should_capture_backtrace=False
        ) as span:
            request._scout_view_span = span

            try:
                response = wrapped(*args, **kwargs)
            except Exception as exc:
                tracked_request.tag("error", "true")
                if scout_config.value("errors_enabled"):
                    ErrorMonitor.send(
                        sys.exc_info(),
                        request_components=get_request_components(self.app, request),
                        request_path=request.path,
                        request_params=dict(request.args.lists()),
                        session=dict(session.items()),
                        environment=self.app.config,
                    )
                raise exc
            else:
                if 500 <= response.status_code <= 599:
                    tracked_request.tag("error", "true")
                return response

    @wrapt.decorator
    def wrapped_preprocess_request(self, wrapped, instance, args, kwargs):
        tracked_request = getattr(request, "_scout_tracked_request", None)
        if tracked_request is None:
            return wrapped(*args, **kwargs)

        # Unlike middleware in other frameworks, using request preprocessors is
        # less common in Flask, so only add a span if there is any in use
        have_before_request_funcs = (
            None in instance.before_request_funcs
            or request.blueprint in instance.before_request_funcs
        )
        if not have_before_request_funcs:
            return wrapped(*args, **kwargs)

        with tracked_request.span("PreprocessRequest", should_capture_backtrace=False):
            return wrapped(*args, **kwargs)

    def extract_flask_settings(self):
        """
        Copies SCOUT_* settings in the app into Scout's config lookup
        """
        configs = {}
        configs["application_root"] = self.app.instance_path
        for name in current_app.config:
            if name.startswith("SCOUT_"):
                value = current_app.config[name]
                clean_name = name.replace("SCOUT_", "").lower()
                configs[clean_name] = value
        scout_config.set(**configs)


def get_request_components(app, request):
    view_func = app.view_functions[request.endpoint]
    request_components = RequestComponents(
        module=view_func.__module__,
        controller=view_func.__name__,
        action=request.method,
    )
    return request_components



# coding=utf-8

import argparse
import logging

from scout_apm.core import CoreAgentManager

logger = logging.getLogger(__name__)


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="count"
    )

    subparsers = parser.add_subparsers(
        title="subcommands", description="valid subcommands", dest="subparser"
    )
    subparsers.add_parser("download")
    subparsers.add_parser("launch")

    args = parser.parse_args(argv)

    if args.verbose is not None:
        if args.verbose >= 2:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)

    core_agent_manager = CoreAgentManager()
    getattr(core_agent_manager, args.subparser)()



# coding=utf-8

import dramatiq

import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest


class ScoutMiddleware(dramatiq.Middleware):
    def __init__(self):
        installed = scout_apm.core.install()
        self._do_nothing = not installed

    def before_process_message(self, broker, message):
        if self._do_nothing:
            return
        tracked_request = TrackedRequest.instance()
        tracked_request.tag("queue", message.queue_name)
        tracked_request.tag("message_id", message.message_id)
        operation = "Job/" + message.actor_name
        tracked_request.start_span(operation=operation)
        tracked_request.operation = operation

    def after_process_message(self, broker, message, result=None, exception=None):
        if self._do_nothing:
            return
        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        if exception:
            tracked_request.tag("error", "true")
        tracked_request.stop_span()

    def after_skip_message(self, broker, message):
        """
        The message was skipped by another middleware raising SkipMessage.
        Stop the span and thus the request, it won't have been marked as real
        so that's alright.
        """
        if self._do_nothing:
            return
        tracked_request = TrackedRequest.instance()
        tracked_request.stop_span()



# coding=utf-8

import wrapt
from flask_sqlalchemy import SQLAlchemy

import scout_apm.sqlalchemy


def instrument_sqlalchemy(db):
    # Version 3 of flask_sqlalchemy changed how engines are created
    if hasattr(db, "_make_engine"):
        db._make_engine = wrapped_make_engine(db._make_engine)
    else:
        SQLAlchemy.get_engine = wrapped_get_engine(SQLAlchemy.get_engine)


@wrapt.decorator
def wrapped_get_engine(wrapped, instance, args, kwargs):
    engine = wrapped(*args, **kwargs)
    scout_apm.sqlalchemy.instrument_sqlalchemy(engine)
    return engine


@wrapt.decorator
def wrapped_make_engine(wrapped, instance, args, kwargs):
    engine = wrapped(*args, **kwargs)
    scout_apm.sqlalchemy.instrument_sqlalchemy(engine)
    return engine



# coding=utf-8

import importlib
import logging

from scout_apm.core.config import scout_config

logger = logging.getLogger(__name__)

instrument_names = ["elasticsearch", "jinja2", "pymongo", "redis", "urllib3"]


def ensure_all_installed():
    disabled_instruments = scout_config.value("disabled_instruments")
    for instrument_name in instrument_names:
        if instrument_name in disabled_instruments:
            logger.info("%s instrument is disabled. Skipping.", instrument_name)
            continue

        module = importlib.import_module("{}.{}".format(__name__, instrument_name))
        module.ensure_installed()



# coding=utf-8

import random
from typing import Dict, Optional, Tuple


class Sampler:
    """
    Handles sampling decision logic for Scout APM.

    This class encapsulates all sampling-related functionality including:
    - Loading and managing sampling configuration
    - Pattern matching for operations (endpoints and jobs)
    - Making sampling decisions based on operation type and patterns
    """

    # Constants for operation type detection
    CONTROLLER_PREFIX = "Controller/"
    JOB_PREFIX = "Job/"

    def __init__(self, config):
        """
        Initialize sampler with Scout configuration.

        Args:
            config: ScoutConfig instance containing sampling configuration
        """
        self.config = config
        self.sample_rate = config.value("sample_rate")
        self.sample_endpoints = config.value("sample_endpoints")
        self.sample_jobs = config.value("sample_jobs")
        self.ignore_endpoints = set(
            config.value("ignore_endpoints") + config.value("ignore")
        )
        self.ignore_jobs = set(config.value("ignore_jobs"))
        self.endpoint_sample_rate = config.value("endpoint_sample_rate")
        self.job_sample_rate = config.value("job_sample_rate")

    def _any_sampling(self):
        """
        Check if any sampling is enabled.

        Returns:
            Boolean indicating if any sampling is enabled
        """
        return (
            self.sample_rate < 1
            or self.sample_endpoints
            or self.sample_jobs
            or self.ignore_endpoints
            or self.ignore_jobs
            or self.endpoint_sample_rate is not None
            or self.job_sample_rate is not None
        )

    def _find_matching_rate(
        self, name: str, patterns: Dict[str, float]
    ) -> Optional[str]:
        """
        Finds the matching sample rate for a given operation name.

        Args:
            name: The operation name to match
            patterns: Dictionary of pattern to sample rate mappings

        Returns:
            The sample rate for the matching pattern or None if no match found
        """

        for pattern, rate in patterns.items():
            if name.startswith(pattern):
                return rate
        return None

    def _get_operation_type_and_name(
        self, operation: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Determines if an operation is an endpoint or job and extracts its name.

        Args:
            operation: The full operation string (e.g. "Controller/users/show")

        Returns:
            Tuple of (type, name) where type is either 'endpoint' or 'job',
            and name is the operation name without the prefix
        """
        if operation.startswith(self.CONTROLLER_PREFIX):
            return "endpoint", operation[len(self.CONTROLLER_PREFIX) :]
        elif operation.startswith(self.JOB_PREFIX):
            return "job", operation[len(self.JOB_PREFIX) :]
        else:
            return None, None

    def get_effective_sample_rate(self, operation: str, is_ignored: bool) -> float:
        """
        Determines the effective sample rate for a given operation.

        Prioritization:
        1. Sampling rate for specific endpoint or job
        2. Specified ignore pattern or flag for operation
        3. Global endpoint or job sample rate
        4. Global sample rate

        Args:
            operation: The operation string (e.g. "Controller/users/show")
            is_ignored: boolean for if the specific transaction is ignored

        Returns:
            Float between 0 and 1 representing sample rate
        """
        op_type, name = self._get_operation_type_and_name(operation)
        patterns = self.sample_endpoints if op_type == "endpoint" else self.sample_jobs
        ignores = self.ignore_endpoints if op_type == "endpoint" else self.ignore_jobs
        default_operation_rate = (
            self.endpoint_sample_rate if op_type == "endpoint" else self.job_sample_rate
        )

        if not op_type or not name:
            return self.sample_rate
        matching_rate = self._find_matching_rate(name, patterns)
        if matching_rate is not None:
            return matching_rate
        for prefix in ignores:
            if name.startswith(prefix) or is_ignored:
                return 0
        if default_operation_rate is not None:
            return default_operation_rate

        # Fall back to global sample rate
        return self.sample_rate

    def should_sample(self, operation: str, is_ignored: bool) -> bool:
        """
        Determines if an operation should be sampled.
        If no sampling is enabled, always return True.

        Args:
            operation: The operation string (e.g. "Controller/users/show"
                   or "Job/mailer")

        Returns:
            Boolean indicating whether to sample this operation
        """
        if not self._any_sampling():
            return True
        return random.random() <= self.get_effective_sample_rate(operation, is_ignored)



# coding=utf-8

import itertools
import os
import sys
import sysconfig
import traceback
import warnings
from logging import getLogger

logger = getLogger(__name__)

# Maximum non-Scout frames to target retrieving
LIMIT = 50
# How many upper frames from inside Scout to ignore
IGNORED = 1


def filter_frames(frames):
    """Filter the stack trace frames down to non-library code."""
    paths = sysconfig.get_paths()
    library_paths = {paths["purelib"], paths["platlib"]}
    for frame in frames:
        if not any(frame["file"].startswith(exclusion) for exclusion in library_paths):
            yield frame


def module_filepath(module, filepath):
    """Get the filepath relative to the base module."""
    root_module_name = module.split(".", 1)[0]
    if root_module_name == module:
        return os.path.basename(filepath)

    module_dir = None
    root_module = sys.modules[root_module_name]
    try:
        if root_module.__file__:
            module_dir = root_module.__file__.rsplit(os.sep, 2)[0]
        elif root_module.__path__ and isinstance(root_module.__path__, (list, tuple)):
            # Default to using the first path specified for the module.
            module_dir = root_module.__path__[0].rsplit(os.sep, 1)[0]
            if len(root_module.__path__) > 1:
                logger.debug(
                    "{} has {} paths. Use the first and ignore the rest.".format(
                        root_module, len(root_module.__path__)
                    )
                )
    except Exception as exc:
        logger.debug(
            "Error processing module {} and filepath {}".format(root_module, filepath),
            exc_info=exc,
        )

    return filepath.split(module_dir, 1)[-1].lstrip(os.sep) if module_dir else filepath


def filepaths(frame):
    """Get the filepath for frame."""
    module = frame.f_globals.get("__name__", None)
    filepath = frame.f_code.co_filename

    if filepath.endswith(".pyc"):
        filepath = filepath[:-1]

    return filepath, (module_filepath(module, filepath) if module else filepath)


def stacktrace_walker(tb):
    """Iterate over each frame of the stack downards for exceptions."""
    for frame, lineno in traceback.walk_tb(tb):
        name = frame.f_code.co_name
        full_path, relative_path = filepaths(frame)
        yield {
            "file": relative_path,
            "full_path": full_path,
            "line": lineno,
            "function": name,
        }


def backtrace_walker():
    """Iterate over each frame of the stack upwards.

    Taken from python3/traceback.ExtractSummary.extract to support
    iterating over the entire stack, but without creating a large
    data structure.
    """
    start_frame = sys._getframe().f_back
    for frame, lineno in traceback.walk_stack(start_frame):
        name = frame.f_code.co_name
        full_path, relative_path = filepaths(frame)
        yield {
            "file": relative_path,
            "full_path": full_path,
            "line": lineno,
            "function": name,
        }


def capture_backtrace():
    walker = filter_frames(backtrace_walker())
    return list(itertools.islice(walker, LIMIT))


def capture_stacktrace(tb):
    walker = stacktrace_walker(tb)
    return list(reversed(list(itertools.islice(walker, LIMIT))))


def capture():
    warnings.warn(
        "capture is deprecated, instead use capture_backtrace instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return capture_backtrace()



# coding=utf-8

import datetime as dt
import logging
import os
import threading

from scout_apm.core.agent.commands import ApplicationEvent
from scout_apm.core.agent.socket import CoreAgentSocketThread
from scout_apm.core.samplers.cpu import Cpu
from scout_apm.core.samplers.memory import Memory
from scout_apm.core.threading import SingletonThread

logger = logging.getLogger(__name__)


class SamplersThread(SingletonThread):
    _instance_lock = threading.Lock()
    _stop_event = threading.Event()

    def run(self):
        logger.debug("Starting Samplers.")
        instances = [Cpu(), Memory()]

        while True:
            for instance in instances:
                event_value = instance.run()
                if event_value is not None:
                    event_type = instance.metric_type + "/" + instance.metric_name
                    event = ApplicationEvent(
                        event_value=event_value,
                        event_type=event_type,
                        timestamp=dt.datetime.now(dt.timezone.utc),
                        source="Pid: " + str(os.getpid()),
                    )
                    CoreAgentSocketThread.send(event)

            should_stop = self._stop_event.wait(timeout=60)
            if should_stop:
                logger.debug("Stopping Samplers.")
                break



# coding=utf-8

import logging
import sys

import wrapt

from scout_apm.core.tracked_request import TrackedRequest

try:
    from jinja2 import Environment
except ImportError:  # pragma: no cover
    Environment = None

try:
    from jinja2 import Template
except ImportError:  # pragma: no cover
    Template = None

# The async_ module can only be shipped on Python 3.6+
try:
    from scout_apm.async_.instruments.jinja2 import wrapped_render_async
except ImportError:
    wrapped_render_async = None


logger = logging.getLogger(__name__)


have_patched_environment_init = False
have_patched_template_render = False
have_patched_template_render_async = False


def ensure_installed():
    global have_patched_template_render

    logger.debug("Instrumenting Jinja2.")

    if Template is None:
        logger.debug("Couldn't import jinja2.Template - probably not installed.")
        return

    instrument_render_async()

    if not have_patched_template_render:
        try:
            Template.render = wrapped_render(Template.render)
        except Exception as exc:
            logger.warning(
                "Failed to instrument jinja2.Template.render: %r", exc, exc_info=exc
            )
        else:
            have_patched_template_render = True


def instrument_render_async():
    global have_patched_environment_init
    global have_patched_template_render_async

    if wrapped_render_async is None:
        return

    if not have_patched_environment_init and not hasattr(Template, "render_async"):
        try:
            Environment.__init__ = wrapped_environment_init_jinja_v2(
                Environment.__init__
            )
        except Exception as exc:
            logger.warning(
                "Failed to instrument jinja2.Environment.__init__: %r",
                exc,
                exc_info=exc,
            )
        else:
            have_patched_environment_init = True
    elif hasattr(Template, "render_async") and not have_patched_template_render_async:
        try:
            Template.render_async = wrapped_render_async(Template.render_async)
        except Exception as exc:
            logger.warning(
                "Failed to instrument jinja2.Template.render_async: %r",
                exc,
                exc_info=exc,
            )
        else:
            have_patched_template_render_async = True


@wrapt.decorator
def wrapped_render(wrapped, instance, args, kwargs):
    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation="Template/Render") as span:
        span.tag("name", instance.name)
        return wrapped(*args, **kwargs)


@wrapt.decorator
def wrapped_environment_init_jinja_v2(wrapped, instance, args, kwargs):
    """
    Delayed wrapping of render_async(), since Template won't have this method
    until after jinja2.asyncsupport is imported, which since Jinja2 2.11.0 is
    done conditionally in Environment.__init__:
    https://github.com/pallets/jinja/issues/765

    This is no longer needed since Jinja2 v3.0.0
    """
    global have_patched_template_render_async
    result = wrapped(*args, **kwargs)

    if (
        wrapped_render_async is not None
        and not have_patched_template_render_async
        and "jinja2.asyncsupport" in sys.modules
    ):
        try:
            Template.render_async = wrapped_render_async(Template.render_async)
        except Exception as exc:
            logger.warning(
                "Failed to instrument jinja2.Template.render_async: %r",
                exc,
                exc_info=exc,
            )
        else:
            have_patched_template_render_async = True

    return result



# coding=utf-8

import logging
from collections import namedtuple

import wrapt

from scout_apm.compat import get_pos_args, unwrap_decorators
from scout_apm.core.tracked_request import TrackedRequest

try:
    from elasticsearch import VERSION as ELASTICSEARCH_VERSION
    from elasticsearch import Elasticsearch
except ImportError:  # pragma: no cover
    Elasticsearch = None
    ELASTICSEARCH_VERSION = (0, 0, 0)

try:
    # Transport was moved to elastic_transport as of v8.0.0
    from elastic_transport import Transport
except ImportError:  # pragma: no cover
    try:
        from elasticsearch import Transport
    except ImportError:  # pragma: no cover
        Transport = None

logger = logging.getLogger(__name__)


def ensure_installed():
    logger.debug("Instrumenting elasticsearch.")

    if Elasticsearch is None or Transport is None:
        if Elasticsearch is None:
            logger.debug(
                "Couldn't import elasticsearch.Elasticsearch "
                "- probably not installed."
            )
        if Transport is None:
            logger.debug(
                "Couldn't import elasticsearch.Transport or "
                "elastic_transport.Transport - probably not installed."
            )
    else:
        ensure_client_instrumented()
        ensure_transport_instrumented()


ClientMethod = namedtuple("ClientMethod", ["name", "takes_index_argument"])

VERSIONED_CLIENT_METHODS = {
    "v7": [
        ClientMethod("bulk", True),
        ClientMethod("clear_scroll", False),
        ClientMethod("close", False),
        ClientMethod("close_point_in_time", False),
        ClientMethod("count", True),
        ClientMethod("create", True),
        ClientMethod("delete", True),
        ClientMethod("delete_by_query", True),
        ClientMethod("delete_by_query_rethrottle", False),
        ClientMethod("delete_script", False),
        ClientMethod("exists", True),
        ClientMethod("exists_source", True),
        ClientMethod("explain", True),
        ClientMethod("field_caps", True),
        ClientMethod("get", True),
        ClientMethod("get_script", False),
        ClientMethod("get_script_context", False),
        ClientMethod("get_script_languages", False),
        ClientMethod("get_source", True),
        ClientMethod("index", True),
        ClientMethod("info", False),
        ClientMethod("mget", True),
        ClientMethod("msearch", True),
        ClientMethod("msearch_template", True),
        ClientMethod("mtermvectors", True),
        ClientMethod("open_point_in_time", True),
        ClientMethod("ping", False),
        ClientMethod("put_script", False),
        ClientMethod("rank_eval", True),
        ClientMethod("reindex", False),
        ClientMethod("reindex_rethrottle", False),
        ClientMethod("render_search_template", False),
        ClientMethod("scripts_painless_execute", False),
        ClientMethod("scroll", False),
        ClientMethod("search", True),
        ClientMethod("search_mvt", True),
        ClientMethod("search_shards", True),
        ClientMethod("search_template", True),
        ClientMethod("termvectors", True),
        ClientMethod("terms_enum", True),
        ClientMethod("update", True),
        ClientMethod("update_by_query", True),
        ClientMethod("update_by_query_rethrottle", False),
    ],
    "v8": [
        ClientMethod("knn_search", True),
        ClientMethod("health_report", False),
    ],
}

CLIENT_METHODS = VERSIONED_CLIENT_METHODS["v7"][:]
if ELASTICSEARCH_VERSION > (8, 0, 0):
    CLIENT_METHODS += VERSIONED_CLIENT_METHODS["v8"]


have_patched_client = False


def ensure_client_instrumented():
    global have_patched_client

    if not have_patched_client:
        instrumented_count = 0
        for name, takes_index_argument in CLIENT_METHODS:
            try:
                method = getattr(Elasticsearch, name)
                if takes_index_argument:
                    wrapped = wrap_client_index_method(method)
                else:
                    wrapped = wrap_client_method(method)
                setattr(Elasticsearch, name, wrapped)
                instrumented_count += 1
            except Exception as exc:
                logger.debug(
                    "Failed to instrument elasticsearch.Elasticsearch.%s: %r",
                    name,
                    exc,
                    exc_info=exc,
                )
        if instrumented_count == 0:
            logger.warning(
                "Failed to instrument any elasticsearch.Elasticsearch methods."
                " Enable debug logs to view root causes."
            )

        have_patched_client = True


@wrapt.decorator
def wrap_client_index_method(wrapped, instance, args, kwargs):
    # elasticsearch-py 7.5.1 changed the order of arguments for client methods,
    # so to be safe we need to inspect the wrapped method's positional
    # arguments to see if we should pull it from there
    if "index" in kwargs:
        index = kwargs["index"]
    else:
        unwrapped = unwrap_decorators(wrapped)
        pos_args = get_pos_args(unwrapped)
        try:
            index_index = pos_args.index("index")
        except ValueError:  # pragma: no cover
            # This guards against the method not accepting an 'index' argument
            # but they all do - for now
            index = ""
        else:
            try:
                index = args[index_index - 1]  # subtract 'self'
            except IndexError:
                index = ""

    if isinstance(index, (list, tuple)):
        index = ",".join(index)
    if index == "":
        index = "Unknown"
    index = index.title()

    camel_name = "".join(c.title() for c in wrapped.__name__.split("_"))
    operation = "Elasticsearch/{}/{}".format(index, camel_name)
    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation=operation, ignore_children=True):
        return wrapped(*args, **kwargs)


@wrapt.decorator
def wrap_client_method(wrapped, instance, args, kwargs):
    camel_name = "".join(c.title() for c in wrapped.__name__.split("_"))
    operation = "Elasticsearch/{}".format(camel_name)
    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation=operation, ignore_children=True):
        return wrapped(*args, **kwargs)


have_patched_transport = False


def ensure_transport_instrumented():
    global have_patched_transport

    if not have_patched_transport:
        try:
            Transport.perform_request = wrapped_perform_request(
                Transport.perform_request
            )
        except Exception as exc:
            logger.warning(
                "Failed to instrument elasticsearch.Transport.perform_request: %r",
                exc,
                exc_info=exc,
            )

    have_patched_transport = True


def _sanitize_name(name):
    try:
        op = name.split("/")[-1]
        op = op[1:]  # chop leading '_' from op
        known_names = (
            "bench",
            "bulk",
            "count",
            "delete_by_query",
            "execute",
            "exists",
            "explain",
            "field_caps",
            "field_stats",
            "health",
            "mget",
            "mlt",
            "mpercolate",
            "msearch",
            "mtermvectors",
            "percolate",
            "pit",
            "query",
            "rank_eval",
            "reindex",
            "script_context",
            "script_language",
            "scroll",
            "search_shards",
            "source",
            "suggest",
            "template",
            "termvectors",
            "terms_enum",
            "update",
            "update_by_query",
            "search",
        )
        if op in known_names:
            return op.title()
        return "Unknown"
    except Exception:
        return "Unknown"


@wrapt.decorator
def wrapped_perform_request(wrapped, instance, args, kwargs):
    try:
        op = _sanitize_name(args[1])
    except IndexError:
        op = "Unknown"

    tracked_request = TrackedRequest.instance()
    with tracked_request.span(
        operation="Elasticsearch/{}".format(op),
        ignore_children=True,
    ):
        return wrapped(*args, **kwargs)



# coding=utf-8

import sys

import scout_apm.core
from scout_apm.compat import ContextDecorator, text
from scout_apm.core.config import ScoutConfig
from scout_apm.core.error import ErrorMonitor
from scout_apm.core.tracked_request import TrackedRequest

# The async_ module can only be shipped on Python 3.6+
try:
    from scout_apm.async_.api import AsyncDecoratorMixin
except ImportError:

    class AsyncDecoratorMixin(object):
        pass


__all__ = [
    "BackgroundTransaction",
    "Config",
    "Context",
    "Error",
    "WebTransaction",
    "install",
    "instrument",
]


class Context(object):
    @classmethod
    def add(self, key, value):
        """Adds context to the currently executing request.

        :key: Any String identifying the request context.
              Example: "user_ip", "plan", "alert_count"
        :value: Any json-serializable type.
              Example: "1.1.1.1", "free", 100
        :returns: nothing.
        """
        TrackedRequest.instance().tag(key, value)


class Config(ScoutConfig):
    pass


install = scout_apm.core.install


def ignore_transaction():
    TrackedRequest.instance().tag("ignore_transaction", True)


class instrument(AsyncDecoratorMixin, ContextDecorator):
    def __init__(self, operation, kind="Custom", tags=None):
        self.operation = text(kind) + "/" + text(operation)
        if tags is None:
            self.tags = {}
        else:
            self.tags = tags

    def __enter__(self):
        tracked_request = TrackedRequest.instance()
        self.span = tracked_request.start_span(operation=self.operation)
        for key, value in self.tags.items():
            self.tag(key, value)
        return self

    def __exit__(self, *exc):
        tracked_request = TrackedRequest.instance()
        tracked_request.stop_span()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class Transaction(AsyncDecoratorMixin, ContextDecorator):
    """
    This Class is not meant to be used directly.
    Use one of the subclasses
    (WebTransaction or BackgroundTransaction)
    """

    def __init__(self, name, tags=None):
        self.name = text(name)
        if tags is None:
            self.tags = {}
        else:
            self.tags = tags

    @classmethod
    def start(cls, kind, name, tags=None):
        operation = text(kind) + "/" + text(name)

        tracked_request = TrackedRequest.instance()
        tracked_request.operation = operation
        tracked_request.is_real_request = True
        span = tracked_request.start_span(
            operation=operation, should_capture_backtrace=False
        )
        if tags is not None:
            for key, value in tags.items():
                tracked_request.tag(key, value)
        return span

    @classmethod
    def stop(cls):
        tracked_request = TrackedRequest.instance()
        tracked_request.stop_span()
        return True

    # __enter__ must be defined by child classes.

    # *exc is any exception raised. Ignore that
    def __exit__(self, *exc):
        WebTransaction.stop()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class WebTransaction(Transaction):
    @classmethod
    def start(cls, name, tags=None):
        super(WebTransaction, cls).start("Controller", text(name), tags)

    def __enter__(self):
        super(WebTransaction, self).start("Controller", self.name, self.tags)


class BackgroundTransaction(Transaction):
    @classmethod
    def start(cls, name, tags=None):
        super(BackgroundTransaction, cls).start("Job", text(name), tags)

    def __enter__(self):
        super(BackgroundTransaction, self).start("Job", self.name, self.tags)


def rename_transaction(name):
    if name is not None:
        tracked_request = TrackedRequest.instance()
        tracked_request.tag("transaction.name", name)


class Error(object):
    @classmethod
    def capture(
        cls,
        exception,
        request_path=None,
        request_params=None,
        session=None,
        custom_controller=None,
        custom_params=None,
    ):
        """
        Capture the exception manually.

        Utilizes sys.exc_info to gather the traceback. This has the side
        effect that if another exception is raised before calling
        ``Error.capture``, the traceback will match the most recently
        raised exception.

        Includes any context added for the TrackedRequest.

        :exception: Any exception.
        :request_path: Any String identifying the relative path of the request.
              Example: "/hello-world/"
        :request_params: Any json-serializable dict representing the
              querystring parameters.
              Example: {"page": 1}
        :session: Any json-serializable dict representing the
              request session.
              Example: {"step": 0}
        :custom_controller: Any String identifying the controller or job.
              Example: "send_email"
        :custom_params: Any json-serializable dict.
              Example: {"to": "scout@test.com", "from": "no-reply@test.com"}
        :returns: nothing.
        """
        if isinstance(exception, Exception):
            exc_info = (exception.__class__, exception, sys.exc_info()[2])
            ErrorMonitor.send(
                exc_info,
                request_path=request_path,
                request_params=request_params,
                session=session,
                custom_controller=custom_controller,
                custom_params=custom_params,
            )



# coding=utf-8

import datetime as dt
import logging
from contextlib import contextmanager
from uuid import uuid4

from scout_apm.core import backtrace, objtrace
from scout_apm.core.agent.commands import BatchCommand
from scout_apm.core.agent.socket import CoreAgentSocketThread
from scout_apm.core.config import scout_config
from scout_apm.core.n_plus_one_tracker import NPlusOneTracker
from scout_apm.core.sampler import Sampler
from scout_apm.core.samplers.memory import get_rss_in_mb
from scout_apm.core.samplers.thread import SamplersThread

logger = logging.getLogger(__name__)


class TrackedRequest(object):
    """
    This is a container which keeps track of all module instances for a single
    request. For convenience they are made available as attributes based on
    their keyname
    """

    _sampler = None

    @classmethod
    def get_sampler(cls):
        if cls._sampler is None:
            cls._sampler = Sampler(scout_config)
        return cls._sampler

    __slots__ = (
        "sampler",
        "request_id",
        "start_time",
        "end_time",
        "active_spans",
        "complete_spans",
        "tags",
        "is_real_request",
        "_memory_start",
        "n_plus_one_tracker",
        "hit_max",
        "sent",
        "operation",
    )

    # Stop adding new spans at this point, to avoid exhausting memory
    MAX_COMPLETE_SPANS = 1500

    @classmethod
    def instance(cls):
        from scout_apm.core.context import context

        return context.get_tracked_request()

    def __init__(self):
        self.request_id = "req-" + str(uuid4())
        self.start_time = dt.datetime.now(dt.timezone.utc)
        self.end_time = None
        self.active_spans = []
        self.complete_spans = []
        self.tags = {}
        self.is_real_request = False
        self._memory_start = get_rss_in_mb()
        self.n_plus_one_tracker = NPlusOneTracker()
        self.hit_max = False
        self.sent = False
        self.operation = None
        logger.debug("Starting request: %s", self.request_id)

    def __repr__(self):
        # Incomplete to avoid TMI
        return "<TrackedRequest(request_id={}, tags={})>".format(
            repr(self.request_id), repr(self.tags)
        )

    def tag(self, key, value):
        if key in self.tags:
            logger.debug(
                "Overwriting previously set tag for request %s: %s",
                self.request_id,
                key,
            )
        self.tags[key] = value

    def start_span(
        self,
        operation,
        ignore=False,
        ignore_children=False,
        should_capture_backtrace=True,
    ):
        parent = self.current_span()
        if parent is not None:
            parent_id = parent.span_id
            if parent.ignore_children:
                ignore = True
                ignore_children = True
        else:
            parent_id = None

        if len(self.complete_spans) >= self.MAX_COMPLETE_SPANS:
            if not self.hit_max:
                logger.debug(
                    "Hit the maximum number of spans, this trace will be incomplete."
                )
                self.hit_max = True
            ignore = True
            ignore_children = True

        new_span = Span(
            request_id=self.request_id,
            operation=operation,
            ignore=ignore,
            ignore_children=ignore_children,
            parent=parent_id,
            should_capture_backtrace=should_capture_backtrace,
        )
        self.active_spans.append(new_span)
        return new_span

    def stop_span(self):
        try:
            stopping_span = self.active_spans.pop()
        except IndexError as exc:
            logger.debug("Exception when stopping span", exc_info=exc)
        else:
            stopping_span.stop()
            if not stopping_span.ignore:
                stopping_span.annotate()
                self.complete_spans.append(stopping_span)

        if len(self.active_spans) == 0:
            self.finish()

    @contextmanager
    def span(self, *args, **kwargs):
        span = self.start_span(*args, **kwargs)
        try:
            yield span
        finally:
            self.stop_span()

    def current_span(self):
        if self.active_spans:
            return self.active_spans[-1]
        else:
            return None

    # Request is done, release any info we have about it.
    def finish(self):
        from scout_apm.core.context import context

        logger.debug("Stopping request: %s", self.request_id)
        if self.end_time is None:
            self.end_time = dt.datetime.now(dt.timezone.utc)

        if self.is_real_request:
            if not self.sent and self.get_sampler().should_sample(
                self.operation, self.is_ignored()
            ):
                self.tag("mem_delta", self._get_mem_delta())
                self.sent = True
                batch_command = BatchCommand.from_tracked_request(self)
                if scout_config.value("log_payload_content"):
                    logger.debug(
                        "Sending request: %s. Payload: %s",
                        self.request_id,
                        batch_command.message(),
                    )
                else:
                    logger.debug("Sending request: %s.", self.request_id)
                CoreAgentSocketThread.send(batch_command)
            SamplersThread.ensure_started()

        details = " ".join(
            "{}={}".format(key, value)
            for key, value in [
                ("start_time", self.start_time),
                ("end_time", self.end_time),
                ("duration", (self.end_time - self.start_time).total_seconds()),
                ("active_spans", len(self.active_spans)),
                ("complete_spans", len(self.complete_spans)),
                ("tags", len(self.tags)),
                ("hit_max", self.hit_max),
                ("is_real_request", self.is_real_request),
                ("sent", self.sent),
            ]
        )
        logger.debug("Request %s %s", self.request_id, details)
        context.clear_tracked_request(self)

    def _get_mem_delta(self):
        current_mem = get_rss_in_mb()
        if current_mem > self._memory_start:
            return current_mem - self._memory_start
        return 0.0

    # A request is ignored if the tag "ignore_transaction" is set to True
    def is_ignored(self):
        return self.tags.get("ignore_transaction", False)


class Span(object):
    __slots__ = (
        "span_id",
        "start_time",
        "end_time",
        "request_id",
        "operation",
        "ignore",
        "ignore_children",
        "parent",
        "tags",
        "start_objtrace_counts",
        "end_objtrace_counts",
        "should_capture_backtrace",
    )

    def __init__(
        self,
        request_id=None,
        operation=None,
        ignore=False,
        ignore_children=False,
        parent=None,
        should_capture_backtrace=True,
    ):
        self.span_id = "span-" + str(uuid4())
        self.start_time = dt.datetime.now(dt.timezone.utc)
        self.end_time = None
        self.request_id = request_id
        self.operation = operation
        self.ignore = ignore
        self.ignore_children = ignore_children
        self.parent = parent
        self.tags = {}
        self.start_objtrace_counts = objtrace.get_counts()
        self.end_objtrace_counts = (0, 0, 0, 0)
        self.should_capture_backtrace = should_capture_backtrace

    def __repr__(self):
        # Incomplete to avoid TMI
        return "<Span(span_id={}, operation={}, ignore={}, tags={})>".format(
            repr(self.span_id), repr(self.operation), repr(self.ignore), repr(self.tags)
        )

    def stop(self):
        self.end_time = dt.datetime.now(dt.timezone.utc)
        self.end_objtrace_counts = objtrace.get_counts()

    def tag(self, key, value):
        if key in self.tags:
            logger.debug(
                "Overwriting previously set tag for span %s: %s", self.span_id, key
            )
        self.tags[key] = value

    # In seconds
    def duration(self):
        if self.end_time is not None:
            return (self.end_time - self.start_time).total_seconds()
        else:
            # Current, running duration
            return (
                dt.datetime.now(tz=dt.timezone.utc) - self.start_time
            ).total_seconds()

    # Add any interesting annotations to the span. Assumes that we are in the
    # process of stopping this span.
    def annotate(self):
        self.add_allocation_tags()
        if not self.should_capture_backtrace:
            return
        slow_threshold = 0.5
        if self.duration() > slow_threshold:
            self.capture_backtrace()

    def add_allocation_tags(self):
        if not objtrace.is_extension:
            logger.debug("Not adding allocation tags: extension not loaded")
            return

        start_allocs = (
            self.start_objtrace_counts[0]
            + self.start_objtrace_counts[1]
            + self.start_objtrace_counts[2]
        )
        end_allocs = (
            self.end_objtrace_counts[0]
            + self.end_objtrace_counts[1]
            + self.end_objtrace_counts[2]
        )

        # If even one of the counters rolled over, we're pretty much
        # guaranteed to have end_allocs be less than start_allocs.
        # This should rarely happen. Max Unsigned Long Long is a big number
        if end_allocs - start_allocs < 0:
            logger.debug(
                "End allocation count smaller than start allocation "
                "count for span %s: start = %d, end = %d",
                self.span_id,
                start_allocs,
                end_allocs,
            )
            return

        self.tag("allocations", end_allocs - start_allocs)
        self.tag("start_allocations", start_allocs)
        self.tag("stop_allocations", end_allocs)

    def capture_backtrace(self):
        # The core-agent will trim the full_path as necessary.
        self.tag(
            "stack",
            [
                {
                    "file": frame["full_path"],
                    "line": frame["line"],
                    "function": frame["function"],
                }
                for frame in backtrace.capture_backtrace()
            ],
        )



# coding=utf-8

import threading


class SingletonThread(threading.Thread):
    _instance = None
    # Copy these variables into subclasses to avoid sharing:
    # (Would use __init_subclass__() but Python 2 doesn't support it and using
    # metaclasses to achieve the same is a lot of faff.)
    # _instance_lock = threading.Lock()
    # _stop_event = threading.Event()

    @classmethod
    def ensure_started(cls):
        instance = cls._instance
        if instance is not None and instance.is_alive():
            # No need to grab the lock
            return
        with cls._instance_lock:
            if cls._instance is None or not cls._instance.is_alive():
                cls._instance = cls()
                cls._instance.start()

    @classmethod
    def ensure_stopped(cls):
        if cls._instance is None:
            # No need to grab the lock
            return
        with cls._instance_lock:
            if cls._instance is None:
                # Nothing to stop
                return
            elif not cls._instance.is_alive():
                # Thread died
                cls._instance = None
                return

            # Signal stopping
            cls._stop_event.set()
            cls._on_stop()
            cls._instance.join()

            cls._instance = None
            cls._stop_event.clear()

    @classmethod
    def _on_stop(cls):
        """
        Hook to allow subclasses to add extra behaviour to stopping.
        """
        pass

    def __init__(self, *args, **kwargs):
        super(SingletonThread, self).__init__(*args, **kwargs)
        self.daemon = True



# coding=utf-8

import logging

import urllib3
import wrapt

from scout_apm.core.config import scout_config
from scout_apm.core.tracked_request import TrackedRequest

try:
    from urllib3 import HTTPConnectionPool
except ImportError:  # pragma: no cover
    HTTPConnectionPool = None

# Try except separately because _url_from_pool is explicitly imported for urllib3 >= 2.
# HTTPConnectionPool is always required.
try:
    from urllib3.connectionpool import _url_from_pool
except ImportError:  # pragma: no cover

    def _url_from_pool(pool, path):
        pass


logger = logging.getLogger(__name__)

have_patched_pool_urlopen = False


def ensure_installed():
    global have_patched_pool_urlopen

    logger.debug("Instrumenting urllib3.")

    if HTTPConnectionPool is None:
        logger.debug(
            "Couldn't import urllib3.HTTPConnectionPool - probably not installed."
        )
        return False
    elif not have_patched_pool_urlopen:
        try:
            HTTPConnectionPool.urlopen = wrapped_urlopen(HTTPConnectionPool.urlopen)
        except Exception as exc:
            logger.warning(
                "Failed to instrument for Urllib3 HTTPConnectionPool.urlopen: %r",
                exc,
                exc_info=exc,
            )
        else:
            have_patched_pool_urlopen = True


@wrapt.decorator
def wrapped_urlopen(wrapped, instance, args, kwargs):
    def _extract_method(method, *args, **kwargs):
        return method

    try:
        method = _extract_method(*args, **kwargs)
    except TypeError:
        method = "Unknown"

    try:
        if int(urllib3.__version__.split(".")[0]) < 2:
            url = str(instance._absolute_url("/"))
        else:
            url = str(_url_from_pool(instance, "/"))
    except Exception:
        logger.exception("Could not get URL for HTTPConnectionPool")
        url = "Unknown"

    # Don't instrument ErrorMonitor calls
    if str(url).startswith(scout_config.value("errors_host")):
        return wrapped(*args, **kwargs)

    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation="HTTP/{}".format(method)) as span:
        span.tag("url", str(url))
        return wrapped(*args, **kwargs)



# coding=utf-8

import datetime as dt
import sys
from os import getpid

from scout_apm.core.agent.commands import ApplicationEvent, format_dt_for_core_agent
from scout_apm.core.agent.socket import CoreAgentSocketThread
from scout_apm.core.config import scout_config


def report_app_metadata():
    CoreAgentSocketThread.send(
        ApplicationEvent(
            event_type="scout.metadata",
            event_value=get_metadata(),
            source="Pid: " + str(getpid()),
            timestamp=dt.datetime.now(dt.timezone.utc),
        )
    )


def get_metadata():
    data = {
        "language": "python",
        "language_version": "{}.{}.{}".format(*sys.version_info[:3]),
        "server_time": format_dt_for_core_agent(dt.datetime.now(dt.timezone.utc)),
        "framework": scout_config.value("framework"),
        "framework_version": scout_config.value("framework_version"),
        "environment": "",
        "app_server": scout_config.value("app_server"),
        "hostname": scout_config.value("hostname"),
        "database_engine": "",
        "database_adapter": "",
        "application_name": "",
        "libraries": get_python_packages_versions(),
        "paas": "",
        "application_root": scout_config.value("application_root"),
        "scm_subdirectory": scout_config.value("scm_subdirectory"),
        "git_sha": scout_config.value("revision_sha"),
    }
    # Deprecated - see #327:
    data["version"] = data["language_version"]
    return data


def get_python_packages_versions():
    try:
        from importlib.metadata import distributions
    except ImportError:
        # For some reason it is unavailable
        return []

    return sorted(
        (
            distribution.metadata["Name"],
            (distribution.metadata["Version"] or "Unknown"),
        )
        for distribution in distributions()
        # Filter out distributions wtih None for name or value. This can be the
        # case for packages without a METADATA or PKG-INFO file in their relevant
        # distribution directory. According to comments in importlib.metadata
        # internals this is possible for certain old packages, but I could only
        # recreate it by deliberately deleting said files.
        if distribution.metadata["Name"]
    )



# coding=utf-8

import datetime as dt
import gzip
import inspect
import queue
from contextlib import ContextDecorator
from functools import wraps
from html import escape
from urllib.parse import parse_qsl, urlencode, urljoin

import certifi
import urllib3


def iteritems(dictionary):
    return dictionary.items()


# datetime_to_timestamp converts a naive UTC datetime to a unix timestamp
def datetime_to_timestamp(datetime_obj):
    return datetime_obj.replace(tzinfo=dt.timezone.utc).timestamp()


def text(value, encoding="utf-8", errors="strict"):
    """
    Convert a value to str on Python 3 and unicode on Python 2.
    """
    if isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        return str(value, encoding, errors)
    else:
        return str(value)


def get_pos_args(func):
    return inspect.getfullargspec(func).args


def unwrap_decorators(func):
    unwrapped = func
    while True:
        # N.B. only some decorators set __wrapped__ on Python 2.7
        try:
            unwrapped = unwrapped.__wrapped__
        except AttributeError:
            break
    return unwrapped


def kwargs_only(func):
    """
    Source: https://pypi.org/project/kwargs-only/
    Make a function only accept keyword arguments.
    This can be dropped in Python 3 in lieu of:
        def foo(*, bar=default):
    Source: https://pypi.org/project/kwargs-only/
    """
    if hasattr(inspect, "signature"):  # pragma: no cover
        # Python 3
        signature = inspect.signature(func)
        arg_names = list(signature.parameters.keys())
    else:  # pragma: no cover
        # Python 2
        signature = inspect.getargspec(func)
        arg_names = signature.args

    if len(arg_names) > 0 and arg_names[0] in ("self", "cls"):
        allowable_args = 1
    else:
        allowable_args = 0

    @wraps(func)
    def wrapper(*args, **kwargs):
        if len(args) > allowable_args:
            raise TypeError(
                "{} should only be called with keyword args".format(func.__name__)
            )
        return func(*args, **kwargs)

    return wrapper


def urllib3_cert_pool_manager(**kwargs):
    return urllib3.PoolManager(cert_reqs="CERT_REQUIRED", ca_certs=certifi.where())


def gzip_compress(data):
    return gzip.compress(data)


__all__ = [
    "ContextDecorator",
    "datetime_to_timestamp",
    "escape",
    "gzip_compress",
    "kwargs_only",
    "parse_qsl",
    "queue",
    "text",
    "urlencode",
    "urljoin",
]



# coding=utf-8

from scout_apm.compat import parse_qsl, urlencode
from scout_apm.core.config import scout_config
from scout_apm.core.queue_time import track_request_queue_time

# Originally derived from:
# 1. Rails:
#   https://github.com/rails/rails/blob/0196551e6039ca864d1eee1e01819fcae12c1dc9/railties/lib/rails/generators/rails/app/templates/config/initializers/filter_parameter_logging.rb.tt  # noqa
# 2. Sentry server side scrubbing:
#   https://docs.sentry.io/data-management/sensitive-data/#server-side-scrubbing
FILTER_PARAMETERS = frozenset(
    [
        "access",
        "access_token",
        "api_key",
        "apikey",
        "auth",
        "auth_token",
        "card[number]",
        "certificate",
        "credentials",
        "crypt",
        "key",
        "mysql_pwd",
        "otp",
        "passwd",
        "password",
        "private",
        "protected",
        "salt",
        "secret",
        "secret_key",
        "ssn",
        "stripetoken",
        "token",
    ]
)


def create_filtered_path(path, query_params):
    if scout_config.value("uri_reporting") == "path":
        return path
    filtered_params = sorted(
        [
            (
                str(key).encode("utf-8"),
                # Apply str again to cover the None case.
                str(filter_element(key, value)).encode("utf-8"),
            )
            for key, value in query_params
        ]
    )
    if not filtered_params:
        return path
    return path + "?" + urlencode(filtered_params)


def filter_element(key, value):
    """
    Filter an individual key/value element of sensitive content. If the
    value is a dictionary, recursively filter the keys in that dictionary.

    Can be used recursively on a dict with:

        filter_element('', {"foo": "bar"})
    """
    is_sensitive = str(key).lower() in FILTER_PARAMETERS

    if is_sensitive:
        filtered = "[FILTERED]"
    elif isinstance(value, dict):
        # Python 2 unicode compatibility: force all keys and values to bytes
        #
        # We expect query_params to have keys and values both of strings, because
        # that's how frameworks build it. However sometimes application code
        # mutates this structure to use incorrect types, or we are filtering a
        # different collection, so we have to cautiously make everything a string
        # again. Ignoring the possibilities of bytes or objects with bad __str__
        # methods because they seem very unlikely.
        filtered = {str(k): filter_element(k, v) for k, v in value.items()}
    elif isinstance(value, list):
        filtered = [filter_element("", v) for v in value]
    elif isinstance(value, set):
        filtered = {filter_element("", v) for v in value}
    elif isinstance(value, tuple):
        filtered = tuple([filter_element("", v) for v in value])
    elif value is None:
        filtered = value
    else:
        filtered = str(value)

    return filtered


def ignore_path(path):
    ignored_paths = scout_config.value("ignore")
    for ignored in ignored_paths:
        if path.lstrip(" /").startswith(ignored):
            return True
    return False


def asgi_track_request_data(scope, tracked_request):
    """
    Track request data from an ASGI HTTP or Websocket scope.
    """
    path = scope.get("root_path", "") + scope["path"]
    query_params = parse_qsl(scope.get("query_string", b"").decode("utf-8"))
    tracked_request.tag("path", create_filtered_path(path, query_params))
    if ignore_path(path):
        tracked_request.tag("ignore_transaction", True)

    # We only care about the last values of headers so don't care that we use
    # a plain dict rather than a multi-value dict
    headers = {k.lower(): v for k, v in scope.get("headers", ())}

    if scout_config.value("collect_remote_ip"):
        user_ip = (
            headers.get(b"x-forwarded-for", b"").decode("latin1").split(",")[0]
            or headers.get(b"client-ip", b"").decode("latin1").split(",")[0]
            or scope.get("client", ("",))[0]
        )
        tracked_request.tag("user_ip", user_ip)

    queue_time = headers.get(b"x-queue-start", b"") or headers.get(
        b"x-request-start", b""
    )
    track_request_queue_time(queue_time.decode("latin1"), tracked_request)


def werkzeug_track_request_data(werkzeug_request, tracked_request):
    """
    Several integrations use Werkzeug requests, so share the code for
    extracting common data here.
    """
    path = werkzeug_request.path
    tracked_request.tag(
        "path", create_filtered_path(path, werkzeug_request.args.items(multi=True))
    )
    if ignore_path(path):
        tracked_request.tag("ignore_transaction", True)

    if scout_config.value("collect_remote_ip"):
        # Determine a remote IP to associate with the request. The value is
        # spoofable by the requester so this is not suitable to use in any
        # security sensitive context.
        user_ip = (
            werkzeug_request.headers.get("x-forwarded-for", default="").split(",")[0]
            or werkzeug_request.headers.get("client-ip", default="").split(",")[0]
            or werkzeug_request.remote_addr
        )
        tracked_request.tag("user_ip", user_ip)

    queue_time = werkzeug_request.headers.get(
        "x-queue-start", default=""
    ) or werkzeug_request.headers.get("x-request-start", default="")
    track_request_queue_time(queue_time, tracked_request)


class RequestComponents(object):
    __slots__ = ("module", "controller", "action")

    def __init__(self, module, controller, action):
        self.module = module
        self.controller = controller
        self.action = action



# coding=utf-8

import json
import logging
import os
import threading
import time

from scout_apm.compat import (
    escape,
    gzip_compress,
    queue,
    urlencode,
    urljoin,
    urllib3_cert_pool_manager,
)
from scout_apm.core.config import scout_config
from scout_apm.core.threading import SingletonThread

# Time unit - monkey-patched in tests to make them run faster
SECOND = 1

logger = logging.getLogger(__name__)


class ErrorServiceThread(SingletonThread):
    _instance_lock = threading.Lock()
    _stop_event = threading.Event()
    _queue = queue.Queue(maxsize=500)

    @classmethod
    def _on_stop(cls):
        super(ErrorServiceThread, cls)._on_stop()
        # Unblock _queue.get()
        try:
            cls._queue.put(None, False)
        except queue.Full as exc:
            logger.debug("ErrorServiceThread full for stop: %r", exc, exc_info=exc)
            pass

    @classmethod
    def send(cls, error):
        try:
            cls._queue.put(error, False)
        except queue.Full as exc:
            logger.debug("ErrorServiceThread full for send: %r", exc, exc_info=exc)

        cls.ensure_started()

    @classmethod
    def wait_until_drained(cls, timeout_seconds=2.0, callback=None):
        interval_seconds = min(timeout_seconds, 0.05)
        start = time.time()
        while True:
            queue_size = cls._queue.qsize()
            queue_empty = queue_size == 0
            elapsed = time.time() - start
            if queue_empty or elapsed >= timeout_seconds:
                break

            if callback is not None:
                callback(queue_size)
                callback = None

            cls.ensure_started()

            time.sleep(interval_seconds)
        return queue_empty

    def run(self):
        batch_size = scout_config.value("errors_batch_size") or 1
        http = urllib3_cert_pool_manager()
        try:
            while True:
                errors = []
                try:
                    # Attempt to fetch the batch size off of the queue.
                    for _ in range(batch_size):
                        error = self._queue.get(block=True, timeout=1 * SECOND)
                        if error:
                            errors.append(error)
                except queue.Empty:
                    pass

                if errors and self._send(http, errors):
                    for _ in range(len(errors)):
                        self._queue.task_done()

                # Check for stop event after each read. This allows opening,
                # sending, and then immediately stopping.
                if self._stop_event.is_set():
                    logger.debug("ErrorServiceThread stopping.")
                    break
        except Exception as exc:
            logger.debug("ErrorServiceThread exception: %r", exc, exc_info=exc)
        finally:
            http.clear()
            logger.debug("ErrorServiceThread stopped.")

    def _send(self, http, errors):
        try:
            data = json.dumps(
                {
                    "notifier": "scout_apm_python",
                    "environment": scout_config.value("environment"),
                    "root": scout_config.value("application_root"),
                    "problems": errors,
                }
            ).encode("utf-8")
        except (ValueError, TypeError) as exc:
            logger.debug(
                "Exception when serializing error message: %r", exc, exc_info=exc
            )
            return False

        params = {
            "key": scout_config.value("key"),
            "name": escape(scout_config.value("name"), quote=False),
        }
        headers = {
            "Agent-Hostname": scout_config.value("hostname") or "",
            "Content-Encoding": "gzip",
            "Content-Type": "application/json",
            "X-Error-Count": "{}".format(len(errors)),
        }

        encoded_args = urlencode(params)
        full_url = urljoin(
            scout_config.value("errors_host"), "apps/error.scout"
        ) + "?{}".format(encoded_args)

        try:
            # urllib3 requires all parameters to be the same type for
            # python 2.7.
            # Since gzip can only return a str, convert all unicode instances
            # to str.
            response = http.request(
                str("POST"),
                str(full_url),
                body=gzip_compress(data),
                headers={str(key): str(value) for key, value in headers.items()},
            )
            if response.status >= 400:
                logger.debug(
                    (
                        "ErrorServiceThread %r response error on _send:"
                        + " %r on PID: %s on thread: %s"
                    ),
                    response.status,
                    response.data,
                    os.getpid(),
                    threading.current_thread(),
                )
                return False
        except Exception as exc:
            logger.debug(
                (
                    "ErrorServiceThread exception on _send:"
                    + " %r on PID: %s on thread: %s"
                ),
                exc,
                os.getpid(),
                threading.current_thread(),
                exc_info=exc,
            )
            return False
        return True



# coding=utf-8

import datetime as dt
import logging

from celery.signals import before_task_publish, task_failure, task_postrun, task_prerun

from scout_apm.core.queue_time import track_job_queue_time

try:
    import django
    from django.views.debug import SafeExceptionReporterFilter

    def get_safe_settings():
        return SafeExceptionReporterFilter().get_safe_settings()

except ImportError:
    # Django not installed
    get_safe_settings = None

import scout_apm.core
from scout_apm.compat import datetime_to_timestamp
from scout_apm.core.config import scout_config
from scout_apm.core.error import ErrorMonitor
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


def before_task_publish_callback(headers=None, properties=None, **kwargs):
    if "scout_task_start" not in headers:
        headers["scout_task_start"] = datetime_to_timestamp(
            dt.datetime.now(dt.timezone.utc)
        )


def task_prerun_callback(task=None, **kwargs):
    tracked_request = TrackedRequest.instance()
    tracked_request.is_real_request = True

    start_time_header = getattr(task.request, "scout_task_start", None)
    track_job_queue_time(start_time_header, tracked_request)

    task_id = getattr(task.request, "id", None)
    if task_id:
        tracked_request.tag("task_id", task_id)
    parent_task_id = getattr(task.request, "parent_id", None)
    if parent_task_id:
        tracked_request.tag("parent_task_id", parent_task_id)

    delivery_info = getattr(task.request, "delivery_info", None)
    if delivery_info:
        tracked_request.tag("is_eager", delivery_info.get("is_eager", False))
        tracked_request.tag("exchange", delivery_info.get("exchange", "unknown"))
        tracked_request.tag("priority", delivery_info.get("priority", "unknown"))
        tracked_request.tag("routing_key", delivery_info.get("routing_key", "unknown"))
        tracked_request.tag("queue", delivery_info.get("queue", "unknown"))

    operation = "Job/" + task.name
    tracked_request.start_span(operation=operation)
    tracked_request.operation = operation


def task_postrun_callback(task=None, **kwargs):
    tracked_request = TrackedRequest.instance()
    tracked_request.stop_span()


def task_failure_callback(
    sender,
    task_id=None,
    exception=None,
    args=None,
    kwargs=None,
    traceback=None,
    einfo=None,
    **remaining,
):
    tracked_request = TrackedRequest.instance()
    tracked_request.tag("error", "true")

    custom_controller = sender.name
    custom_params = {
        "celery": {
            "task_id": task_id,
            "args": args,
            "kwargs": kwargs,
        }
    }

    # Look up the django settings if populated.
    environment = None
    if get_safe_settings:
        try:
            environment = get_safe_settings()
        except django.core.exceptions.ImproperlyConfigured as exc:
            # Django not setup correctly
            logger.debug(
                "Celery integration does not have django configured properly: %r", exc
            )
            pass
        except Exception as exc:
            logger.debug(
                "Celery task_failure callback exception: %r", exc, exc_info=exc
            )
            pass

    # Celery occassionally will send the traceback as a string rather
    # than a Stack trace object as the docs indicate. In that case,
    # fall back to the billiard ExceptionInfo instance
    traceback = traceback if traceback and not isinstance(traceback, str) else einfo.tb
    exc_info = (exception.__class__, exception, traceback)
    ErrorMonitor.send(
        exc_info,
        environment=environment,
        custom_params=custom_params,
        custom_controller=custom_controller,
    )


def install(app=None):
    if app is not None:
        copy_configuration(app)

    installed = scout_apm.core.install()
    if not installed:
        return

    before_task_publish.connect(before_task_publish_callback)
    task_prerun.connect(task_prerun_callback)
    task_failure.connect(task_failure_callback)
    task_postrun.connect(task_postrun_callback)


def copy_configuration(app):
    prefix = "scout_"
    prefix_len = len(prefix)

    to_set = {}
    for key, value in app.conf.items():
        key_lower = key.lower()
        if key_lower.startswith(prefix) and len(key_lower) > prefix_len:
            scout_key = key_lower[prefix_len:]
            to_set[scout_key] = value

    scout_config.set(**to_set)


def uninstall():
    before_task_publish.disconnect(before_task_publish_callback)
    task_prerun.disconnect(task_prerun_callback)
    task_postrun.disconnect(task_postrun_callback)
    task_failure.disconnect(task_failure_callback)



# coding=utf-8

from scout_apm.core.web_requests import RequestComponents


def get_controller_name(request):
    view_func = request.resolver_match.func
    view_name = request.resolver_match._func_path
    if hasattr(view_func, "view_class"):
        view_func = view_func.view_class
        view_name = "{}.{}".format(view_func.__module__, view_func.__name__)

    django_admin_components = _get_django_admin_components(view_func)
    if django_admin_components:
        view_name = "{}.{}.{}".format(
            django_admin_components.module,
            django_admin_components.controller,
            django_admin_components.action,
        )

    django_rest_framework_components = _get_django_rest_framework_components(
        request, view_func
    )
    if django_rest_framework_components is not None:
        view_name = "{}.{}.{}".format(
            django_rest_framework_components.module,
            django_rest_framework_components.controller,
            django_rest_framework_components.action,
        )

    # Seems to be a Tastypie Resource. Need to resort to some stack inspection
    # to find a better name since its decorators don't wrap very well
    if view_name == "tastypie.resources.wrapper":
        tastypie_components = _get_tastypie_components(request, view_func)
        if tastypie_components is not None:
            view_name = "{}.{}.{}".format(
                tastypie_components.module,
                tastypie_components.controller,
                tastypie_components.action,
            )

    return "Controller/{}".format(view_name)


def get_request_components(request):
    if not request.resolver_match:
        return None
    view_func = request.resolver_match.func
    view_name = request.resolver_match._func_path
    if hasattr(view_func, "view_class"):
        view_func = view_func.view_class
    request_components = RequestComponents(
        module=view_func.__module__,
        controller=view_func.__name__,
        action=request.method,
    )

    django_admin_components = _get_django_admin_components(view_func)
    if django_admin_components:
        request_components = django_admin_components

    django_rest_framework_components = _get_django_rest_framework_components(
        request, view_func
    )
    if django_rest_framework_components is not None:
        request_components = django_rest_framework_components

    # Seems to be a Tastypie Resource. Need to resort to some stack inspection
    # to find a better name since its decorators don't wrap very well
    if view_name == "tastypie.resources.wrapper":
        tastypie_components = _get_tastypie_components(request, view_func)
        if tastypie_components is not None:
            request_components = tastypie_components
    return request_components


def _get_django_admin_components(view_func):
    if hasattr(view_func, "model_admin"):
        # Seems to comes from Django admin (attribute only set on Django 1.9+)
        admin_class = view_func.model_admin.__class__
        return RequestComponents(
            module=admin_class.__module__,
            controller=admin_class.__name__,
            action=view_func.__name__,
        )
    return None


def _get_django_rest_framework_components(request, view_func):
    try:
        from rest_framework.viewsets import ViewSetMixin
    except ImportError:
        return None

    kls = getattr(view_func, "cls", None)
    if isinstance(kls, type) and not issubclass(kls, ViewSetMixin):
        return None

    # Get 'actions' set in ViewSetMixin.as_view
    actions = getattr(view_func, "actions", None)
    if not actions or not isinstance(actions, dict):
        return None

    method_lower = request.method.lower()
    if method_lower not in actions:
        return None

    return RequestComponents(
        module=view_func.__module__,
        controller=view_func.__name__,
        action=actions[method_lower],
    )


def _get_tastypie_components(request, view_func):
    try:
        from tastypie.resources import Resource
    except ImportError:
        return None

    try:
        wrapper = view_func.__wrapped__
    except AttributeError:
        return None

    if not hasattr(wrapper, "__closure__") or len(wrapper.__closure__) != 2:
        return None

    instance = wrapper.__closure__[0].cell_contents
    if not isinstance(instance, Resource):  # pragma: no cover
        return None

    method_name = wrapper.__closure__[1].cell_contents
    if not isinstance(method_name, str):  # pragma: no cover
        return None

    if method_name.startswith("dispatch_"):  # pragma: no cover
        method_name = request.method.lower() + method_name.split("dispatch", 1)[1]

    return RequestComponents(
        module=instance.__module__,
        controller=instance.__class__.__name__,
        action=method_name,
    )



# coding=utf-8

import datetime as dt
import logging

import psutil

logger = logging.getLogger(__name__)


class Cpu(object):
    metric_type = "CPU"
    metric_name = "Utilization"
    human_name = "Process CPU"

    def __init__(self):
        self.last_run = dt.datetime.now(dt.timezone.utc)
        self.last_cpu_times = psutil.Process().cpu_times()
        self.num_processors = psutil.cpu_count()
        if self.num_processors is None:
            logger.debug("Could not determine CPU count - assuming there is one.")
            self.num_processors = 1

    def run(self):
        now = dt.datetime.now(dt.timezone.utc)
        process = psutil.Process()  # get a handle on the current process
        cpu_times = process.cpu_times()

        wall_clock_elapsed = (now - self.last_run).total_seconds()
        if wall_clock_elapsed < 0:
            self.save_times(now, cpu_times)
            logger.debug(
                "%s: Negative time elapsed. now: %s, last_run: %s.",
                self.human_name,
                now,
                self.last_run,
            )
            return None

        utime_elapsed = cpu_times.user - self.last_cpu_times.user
        stime_elapsed = cpu_times.system - self.last_cpu_times.system
        process_elapsed = utime_elapsed + stime_elapsed

        # This can happen right after a fork.  This class starts up in
        # pre-fork, records {u,s}time, then forks. This resets {u,s}time to 0
        if process_elapsed < 0:
            self.save_times(now, cpu_times)
            logger.debug(
                "%s: Negative process time elapsed. "
                "utime: %s, stime: %s, total time: %s. "
                "This is normal to see when starting a forking web server.",
                self.human_name,
                utime_elapsed,
                stime_elapsed,
                process_elapsed,
            )
            return None

        # Normalized to # of processors
        normalized_wall_clock_elapsed = wall_clock_elapsed * self.num_processors

        # If somehow we run for 0 seconds between calls, don't try to divide by 0
        if normalized_wall_clock_elapsed == 0:
            res = 0
        else:
            res = (process_elapsed / normalized_wall_clock_elapsed) * 100

        self.save_times(now, cpu_times)

        logger.debug("%s: %s [%s CPU(s)]", self.human_name, res, self.num_processors)

        return res

    def save_times(self, now, cpu_times):
        self.last_run = now
        self.cpu_times = cpu_times



# coding=utf-8

import logging

from django.template import Template
from django.template.loader_tags import BlockNode

from scout_apm.core.stacktracer import trace_method

logger = logging.getLogger(__name__)

templates_instrumented = False


def ensure_templates_instrumented():
    global templates_instrumented
    if templates_instrumented:
        return
    templates_instrumented = True

    @trace_method(Template)
    def __init__(self, *args, **kwargs):
        name = args[2] if len(args) >= 3 else "<Unknown Template>"
        return ("Template/Compile", {"name": name})

    @trace_method(Template)
    def render(self, *args, **kwargs):
        name = self.name if self.name is not None else "<Unknown Template>"
        return ("Template/Render", {"name": name})

    @trace_method(BlockNode, "render")
    def render_block(self, *args, **kwargs):
        return ("Block/Render", {"name": self.name})

    logger.debug("Monkey patched Templates")



# coding=utf-8

import logging

import wrapt

from scout_apm.core.tracked_request import TrackedRequest

try:
    import pymongo
    from pymongo.collection import Collection
except ImportError:
    pymongo = None
    Collection = None

logger = logging.getLogger(__name__)

have_patched_collection = False


def ensure_installed():
    global have_patched_collection

    logger.debug("Instrumenting pymongo.")

    if Collection is None:
        logger.debug("Couldn't import pymongo.Collection - probably not installed.")
    elif not have_patched_collection:
        methods = COLLECTION_METHODS
        if pymongo.version_tuple < (4, 0):
            methods = COLLECTION_METHODS_V3
        for name in methods:
            try:
                setattr(
                    Collection, name, wrap_collection_method(getattr(Collection, name))
                )
            except Exception as exc:
                logger.warning(
                    "Failed to instrument pymongo.Collection.%s: %r",
                    name,
                    exc,
                    exc_info=exc,
                )
        have_patched_collection = True


COLLECTION_METHODS = [
    "aggregate",
    "aggregate_raw_batches",
    "bulk_write",
    "count_documents",
    "create_index",
    "create_indexes",
    "delete_many",
    "delete_one",
    "distinct",
    "drop",
    "drop_index",
    "drop_indexes",
    "estimated_document_count",
    "find",
    "find_one",
    "find_one_and_delete",
    "find_one_and_replace",
    "find_one_and_update",
    "find_raw_batches",
    "index_information",
    "insert_many",
    "insert_one",
    "list_indexes",
    "rename",
    "replace_one",
    "update_many",
    "update_one",
    "drop_search_index",
    "create_search_indexes",
    "create_search_index",
    "list_search_indexes",
    "update_search_index",
]

COLLECTION_METHODS_V3 = COLLECTION_METHODS + [
    "count",
    "ensure_index",
    "find_and_modify",
    "group",
    "inline_map_reduce",
    "insert",
    "map_reduce",
    "parallel_scan",
    "reindex",
    "remove",
    "save",
    "update",
]


@wrapt.decorator
def wrap_collection_method(wrapped, instance, args, kwargs):
    tracked_request = TrackedRequest.instance()
    camel_name = "".join(c.title() for c in wrapped.__name__.split("_"))
    operation = "MongoDB/{}.{}".format(instance.name, camel_name)
    with tracked_request.span(operation=operation, ignore_children=True) as span:
        span.tag("name", instance.name)
        return wrapped(*args, **kwargs)



# coding=utf-8

import datetime as dt
import logging

import wrapt
from rq import SimpleWorker as RqSimpleWorker
from rq import Worker as RqWorker
from rq.job import Job
from rq.worker import HerokuWorker as RqHerokuWorker

import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest

install_attempted = False
installed = None

logger = logging.getLogger(__name__)


def ensure_scout_installed():
    global install_attempted, installed

    if not install_attempted:
        install_attempted = True
        installed = scout_apm.core.install()


class WorkerMixin(object):
    def __init__(self, *args, **kwargs):
        global installed
        ensure_scout_installed()
        if installed:
            ensure_job_instrumented()
        super(WorkerMixin, self).__init__(*args, **kwargs)


class Worker(WorkerMixin, RqWorker):
    pass


class SimpleWorker(WorkerMixin, RqSimpleWorker):
    pass


class HerokuWorker(WorkerMixin, RqHerokuWorker):
    pass


job_instrumented = False


def ensure_job_instrumented():
    global job_instrumented
    if job_instrumented:
        return
    job_instrumented = True
    Job.perform = wrap_perform(Job.perform)


@wrapt.decorator
def wrap_perform(wrapped, instance, args, kwargs):
    global installed
    if not installed:
        return wrapped(*args, **kwargs)

    tracked_request = TrackedRequest.instance()
    tracked_request.is_real_request = True
    tracked_request.tag("task_id", instance.get_id())
    tracked_request.tag("queue", instance.origin)
    # rq strips tzinfo from enqueued_at during serde in at least some cases
    # internally everything uses UTC naive datetimes, so we operate on that
    # assumption here.
    if instance.enqueued_at.tzinfo is None:
        queued_at = instance.enqueued_at.replace(tzinfo=dt.timezone.utc)
    else:
        queued_at = instance.enqueued_at
    queue_time = (dt.datetime.now(dt.timezone.utc) - queued_at).total_seconds()
    tracked_request.tag("queue_time", queue_time)
    operation = "Job/{}".format(instance.func_name)
    tracked_request.operation = operation
    with tracked_request.span(operation=operation):
        try:
            return wrapped(*args, **kwargs)
        except Exception:
            tracked_request.tag("error", "true")
            raise



# coding=utf-8

import wrapt

from scout_apm.core.tracked_request import TrackedRequest


def trace_method(cls, method_name=None):
    def decorator(info_func):
        method_to_patch = method_name or info_func.__name__

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            entry_type, detail = info_func(instance, *args, **kwargs)
            operation = entry_type
            if detail["name"] is not None:
                operation = operation + "/" + detail["name"]

            tracked_request = TrackedRequest.instance()
            with tracked_request.span(operation=operation) as span:
                for key, value in detail.items():
                    span.tag(key, value)

                return wrapped(*args, **kwargs)

        setattr(cls, method_to_patch, wrapper(getattr(cls, method_to_patch)))

        return wrapper

    return decorator



# coding=utf-8

import logging

from django.conf import settings

logger = logging.getLogger(__name__)

huey_instrumented = False


def ensure_huey_instrumented():
    global huey_instrumented
    if huey_instrumented:
        return
    huey_instrumented = True

    # Avoid importing if not installed
    if "huey.contrib.djhuey" not in settings.INSTALLED_APPS:  # pragma: no cover
        return

    try:
        from huey.contrib.djhuey import HUEY
    except ImportError:  # pragma: no cover
        return

    from scout_apm.huey import attach_scout_handlers

    attach_scout_handlers(HUEY)
    logger.debug("Instrumented huey.contrib.djhuey")



# coding=utf-8

import datetime as dt
import logging
import time
import typing

from scout_apm.compat import datetime_to_timestamp
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)

# Cutoff epoch is used for determining ambiguous timestamp boundaries
CUTOFF_EPOCH_S = time.mktime((dt.date.today().year - 10, 1, 1, 0, 0, 0, 0, 0, 0))
CUTOFF_EPOCH_MS = CUTOFF_EPOCH_S * 1000.0
CUTOFF_EPOCH_US = CUTOFF_EPOCH_S * 1000000.0
CUTOFF_EPOCH_NS = CUTOFF_EPOCH_S * 1000000000.0


def _convert_ambiguous_timestamp_to_ns(timestamp: float) -> float:
    """
    Convert an ambiguous float timestamp that could be in nanoseconds,
    microseconds, milliseconds, or seconds to nanoseconds. Return 0.0 for
    values in the more than 10 years ago.
    """
    if timestamp > CUTOFF_EPOCH_NS:
        converted_timestamp = timestamp
    elif timestamp > CUTOFF_EPOCH_US:
        converted_timestamp = timestamp * 1000.0
    elif timestamp > CUTOFF_EPOCH_MS:
        converted_timestamp = timestamp * 1000000.0
    elif timestamp > CUTOFF_EPOCH_S:
        converted_timestamp = timestamp * 1000000000.0
    else:
        return 0.0
    return converted_timestamp


def track_request_queue_time(
    header_value: typing.Any, tracked_request: TrackedRequest
) -> bool:
    """
    Attempt to parse a queue time header and store the result in the tracked request.

    Returns:
        bool: Whether we succeeded in marking queue time. Used for testing.
    """
    if header_value.startswith("t="):
        header_value = header_value[2:]

    try:
        first_char = header_value[0]
    except IndexError:
        return False

    if not first_char.isdigit():  # filter out negatives, nan, inf, etc.
        return False

    try:
        ambiguous_start_timestamp = float(header_value)
    except ValueError:
        return False

    start_timestamp_ns = _convert_ambiguous_timestamp_to_ns(ambiguous_start_timestamp)
    if start_timestamp_ns == 0.0:
        return False

    tr_start_timestamp_ns = datetime_to_timestamp(tracked_request.start_time) * 1e9

    # Ignore if in the future
    if start_timestamp_ns > tr_start_timestamp_ns:
        return False

    queue_time_ns = int(tr_start_timestamp_ns - start_timestamp_ns)
    tracked_request.tag("scout.queue_time_ns", queue_time_ns)
    return True


def track_job_queue_time(
    header_value: typing.Any, tracked_request: TrackedRequest
) -> bool:
    """
    Attempt to parse a queue/latency time header and store the result in the request.

    Returns:
        bool: Whether we succeeded in marking queue time for the job. Used for testing.
    """
    if header_value is not None:
        now = datetime_to_timestamp(dt.datetime.now(dt.timezone.utc)) * 1e9
        try:
            ambiguous_float_start = typing.cast(float, header_value)
            start = _convert_ambiguous_timestamp_to_ns(ambiguous_float_start)
            queue_time_ns = int(now - start)
        except TypeError:
            logger.debug("Invalid job queue time header: %r", header_value)
            return False
        else:
            tracked_request.tag("scout.job_queue_time_ns", queue_time_ns)
            return True



# coding=utf-8

import logging

import django
import wrapt
from django.db import connections
from django.db.backends.signals import connection_created
from django.db.backends.utils import CursorWrapper

from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)

sql_instrumented = False


def ensure_sql_instrumented():
    global sql_instrumented
    if sql_instrumented:
        return
    sql_instrumented = True

    if django.VERSION >= (2, 0):
        for connection in connections.all():
            install_db_execute_hook(connection=connection)
        connection_created.connect(install_db_execute_hook)
        logger.debug("Installed DB connection created signal handler")
    else:
        CursorWrapper.execute = execute_wrapper(CursorWrapper.execute)
        CursorWrapper.executemany = executemany_wrapper(CursorWrapper.executemany)

        logger.debug("Monkey patched SQL")


def db_execute_hook(execute, sql, params, many, context):
    """
    Database instrumentation hook for Django 2.0+
    https://docs.djangoproject.com/en/2.0/topics/db/instrumentation/
    """
    if many:
        operation = "SQL/Many"
    else:
        operation = "SQL/Query"

    if sql is not None:
        tracked_request = TrackedRequest.instance()
        span = tracked_request.start_span(operation=operation)
        span.tag("db.statement", sql)

    try:
        return execute(sql, params, many, context)
    finally:
        if sql is not None:
            tracked_request.stop_span()
            if tracked_request.n_plus_one_tracker.should_capture_backtrace(
                sql=sql,
                duration=span.duration(),
                count=(1 if not many else len(params)),
            ):
                span.capture_backtrace()


def install_db_execute_hook(connection, **kwargs):
    """
    Install db_execute_hook on the given database connection.

    Rather than use the documented API of the `execute_wrapper()` context
    manager, directly insert the hook. This is done because:
    1. Deleting the context manager closes it, so it's not possible to enter
       it here and not exit it, unless we store it forever in some variable
    2. We want to be the first hook, so we can capture every query (although
       potentially later hooks will change the SQL)
    3. We want to be idempotent and only install the hook once
    """
    if db_execute_hook not in connection.execute_wrappers:
        connection.execute_wrappers.insert(0, db_execute_hook)


@wrapt.decorator
def execute_wrapper(wrapped, instance, args, kwargs):
    """
    CursorWrapper.execute() wrapper for Django < 2.0
    """
    try:
        sql = _extract_sql(*args, **kwargs)
    except TypeError:
        sql = None

    if sql is not None:
        tracked_request = TrackedRequest.instance()
        span = tracked_request.start_span(operation="SQL/Query")
        span.tag("db.statement", sql)

    try:
        return wrapped(*args, **kwargs)
    finally:
        if sql is not None:
            tracked_request.stop_span()
            if tracked_request.n_plus_one_tracker.should_capture_backtrace(
                sql, span.duration()
            ):
                span.capture_backtrace()


def _extract_sql(sql, *args, **kwargs):
    return sql


@wrapt.decorator
def executemany_wrapper(wrapped, instance, args, kwargs):
    """
    CursorWrapper.executemany() wrapper for Django < 2.0
    """
    try:
        sql, param_list = _extract_sql_param_list(*args, **kwargs)
    except TypeError:
        sql = None
        param_list = None

    if sql is not None:
        tracked_request = TrackedRequest.instance()
        span = tracked_request.start_span(operation="SQL/Many")
        span.tag("db.statement", sql)

    try:
        return wrapped(*args, **kwargs)
    finally:
        if sql is not None:
            tracked_request.stop_span()
            if tracked_request.n_plus_one_tracker.should_capture_backtrace(
                sql=sql,
                duration=span.duration(),
                count=len(param_list),
            ):
                span.capture_backtrace()


def _extract_sql_param_list(sql, param_list, *args, **kwargs):
    return sql, param_list



# coding=utf-8

import logging
import os

from scout_apm.core.backtrace import capture_stacktrace
from scout_apm.core.config import scout_config
from scout_apm.core.error_service import ErrorServiceThread
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import RequestComponents, filter_element

logger = logging.getLogger(__name__)


class ErrorMonitor(object):
    @classmethod
    def send(
        cls,
        exc_info,
        request_components=None,
        request_path=None,
        request_params=None,
        session=None,
        environment=None,
        custom_controller=None,
        custom_params=None,
    ):
        if not scout_config.value("errors_enabled"):
            return

        exc_class, exc_value, traceback = exc_info

        ignore_exceptions = scout_config.value("errors_ignored_exceptions")
        if ignore_exceptions and isinstance(exc_value, tuple(ignore_exceptions)):
            return

        tracked_request = TrackedRequest.instance()

        # If monitoring is disabled, we need to clean up the tracked request
        # to avoid leaking it (since the middleware won't finish it)
        if not scout_config.value("monitor"):
            # Finish the request immediately to prevent leaks
            tracked_request.finish()
            return

        context = {}
        context.update(tracked_request.tags)
        context["transaction_id"] = tracked_request.request_id

        if custom_params:
            context["custom_params"] = custom_params

        if custom_controller:
            if request_components:
                request_components.controller = custom_controller
            else:
                request_components = RequestComponents(
                    module=None, controller=custom_controller, action=None
                )

        scm_subdirectory = scout_config.value("scm_subdirectory")
        error = {
            "exception_class": exc_class.__name__,
            "message": str(exc_value),
            "request_id": tracked_request.request_id,
            "request_uri": request_path,
            "request_params": (
                filter_element("", request_params) if request_params else None
            ),
            "request_session": filter_element("", session) if session else None,
            "environment": filter_element("", environment) if environment else None,
            "trace": [
                "{file}:{line}:in {function}".format(
                    file=(
                        os.path.join(scm_subdirectory, frame["file"])
                        if scm_subdirectory
                        else frame["file"]
                    ),
                    line=frame["line"],
                    function=frame["function"],
                )
                for frame in capture_stacktrace(traceback)
            ],
            "request_components": (
                {
                    "module": request_components.module,
                    "controller": request_components.controller,
                    "action": request_components.action,
                }
                if request_components
                else None
            ),
            "context": context,
            "host": scout_config.value("hostname"),
            "revision_sha": scout_config.value("revision_sha"),
        }

        if scout_config.value("log_payload_content"):
            logger.debug(
                "Sending error for request: %s. Payload: %r",
                tracked_request.request_id,
                error,
            )
        else:
            logger.debug("Sending error for request: %s.", tracked_request.request_id)

        ErrorServiceThread.send(error=error)



# coding=utf-8

# Try use the C extension, but if it isn't available, provide a dummy
# implementation.

try:
    from scout_apm.core._objtrace import disable, enable, get_counts, reset_counts
except ImportError:  # pragma: no cover

    def enable():
        pass

    def disable():
        pass

    def get_counts():
        return (0, 0, 0, 0)

    def reset_counts():
        pass

    is_extension = False
else:  # pragma: no cover
    is_extension = True



# coding=utf-8

import wrapt

from scout_apm.core.tracked_request import TrackedRequest


@wrapt.decorator
async def wrapped_render_async(wrapped, instance, args, kwargs):
    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation="Template/Render") as span:
        span.tag("name", instance.name)
        return await wrapped(*args, **kwargs)



# coding=utf-8

import hug
from hug.interface import HTTP

from scout_apm.falcon import ScoutMiddleware as FalconMiddleware


class ScoutMiddleware(FalconMiddleware):
    """
    Hug's HTTP interface is based on Falcon. Therefore we use a subclass of our
    Falcon integration with Hug specific extras.
    """

    def __init__(self, config, hug_http_interface):
        super(ScoutMiddleware, self).__init__(config)
        self.hug_http_interface = hug_http_interface

    def process_request(self, req, resp):
        if not self._do_nothing and self.api is None:
            self.api = self.hug_http_interface.falcon
        return super(ScoutMiddleware, self).process_request(req, resp)

    def _name_operation(self, req, responder, resource):
        if isinstance(responder, HTTP):
            # Hug doesn't use functions but its custom callable classes
            return "Controller/{}.{}".format(
                responder.interface._function.__module__,
                responder.interface._function.__name__,
            )
        return super(ScoutMiddleware, self)._name_operation(req, responder, resource)


def integrate_scout(hug_module_name, config):
    http_interface = hug.API(hug_module_name).http
    scout_middleware = ScoutMiddleware(
        config=config,
        hug_http_interface=http_interface,
    )
    http_interface.add_middleware(scout_middleware)



# coding=utf-8

from collections import defaultdict


class NPlusOneTrackedItem(object):
    __slots__ = ("count", "duration", "captured")

    def __init__(self):
        self.count = 0
        self.duration = 0.0
        self.captured = False


class NPlusOneTracker(object):
    # Fetch backtraces on this number of same SQL calls
    COUNT_THRESHOLD = 5

    # Minimum time in seconds before we start performing any work.
    DURATION_THRESHOLD = 0.150

    __slots__ = ("_map",)

    def __init__(self):
        self._map = defaultdict(NPlusOneTrackedItem)

    def should_capture_backtrace(self, sql, duration, count=1):
        item = self._map[sql]
        if item.captured:
            return False

        item.duration += duration
        item.count += count

        if (
            item.duration >= self.DURATION_THRESHOLD
            and item.count >= self.COUNT_THRESHOLD
        ):
            item.captured = True
            return True
        return False



# coding=utf-8

import platform


def is_valid_triple(triple):
    values = triple.split("-", 1)
    return (
        len(values) == 2
        and values[0] in ("i686", "x86_64", "aarch64", "unknown")
        and values[1]
        in ("unknown-linux-gnu", "unknown-linux-musl", "apple-darwin", "unknown")
        # Validate _apple_darwin_aarch64_override was applied.
        and triple != "aarch64-apple-darwin"
    )


def _apple_darwin_aarch64_override(triple):
    """
    If using M1 ARM64 machine, still use x86_64.
    See https://github.com/scoutapp/scout_apm_python/issues/683
    """
    if triple == "aarch64-apple-darwin":
        return "x86_64-apple-darwin"
    return triple


def get_triple():
    return _apple_darwin_aarch64_override(
        "{arch}-{platform}".format(arch=get_arch(), platform=get_platform())
    )


def get_arch():
    """
    What CPU are we on?
    """
    arch = platform.machine()
    if arch == "i686":
        return "i686"
    elif arch == "x86_64":
        return "x86_64"
    elif arch == "aarch64":
        return "aarch64"
    elif arch == "arm64":
        # We treat these as synonymous and name them "aarch64" for consistency
        # Mac OS, for example, uses "arm64"
        return "aarch64"
    else:
        return "unknown"


def get_platform():
    """
    What Operating System (and sub-system like glibc / musl)
    """
    system_name = platform.system()
    if system_name == "Linux":
        # Previously we'd use either "-gnu" or "-musl" indicate which version
        # of libc we were built against. We now default to musl since it
        # reliably works on all platforms.
        return "unknown-linux-musl"
    elif system_name == "Darwin":
        return "apple-darwin"
    else:
        return "unknown"



# coding=utf-8

from huey.exceptions import RetryTask, TaskLockedException
from huey.signals import SIGNAL_CANCELED

import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest

# Because neither hooks nor signals are called in *all* cases, we need to use
# both in order to capture every case. See source:
# https://github.com/coleifer/huey/blob/e6710bd6a9f581ebc728e24f5923d26eb0047750/huey/api.py#L331  # noqa


def attach_scout(huey):
    installed = scout_apm.core.install()
    if installed:
        attach_scout_handlers(huey)


def attach_scout_handlers(huey):
    huey.pre_execute()(scout_on_pre_execute)
    huey.post_execute()(scout_on_post_execute)
    huey.signal(SIGNAL_CANCELED)(scout_on_cancelled)


def scout_on_pre_execute(task):
    tracked_request = TrackedRequest.instance()

    tracked_request.tag("task_id", task.id)

    operation = "Job/{}.{}".format(task.__module__, task.__class__.__name__)
    tracked_request.start_span(operation=operation)
    tracked_request.operation = operation


def scout_on_post_execute(task, task_value, exception):
    tracked_request = TrackedRequest.instance()
    if exception is None:
        tracked_request.is_real_request = True
    elif isinstance(exception, TaskLockedException):
        pass
    elif isinstance(exception, RetryTask):
        tracked_request.is_real_request = True
        tracked_request.tag("retrying", True)
    else:
        tracked_request.is_real_request = True
        tracked_request.tag("error", "true")
    tracked_request.stop_span()


def scout_on_cancelled(signal, task, exc=None):
    # In the case of a cancelled signal, Huey doesn't run the post_execute
    # handler, so we need to tidy up
    TrackedRequest.instance().stop_span()



# coding=utf-8

import atexit
import logging
import os
import sys

from scout_apm import instruments
from scout_apm.compat import kwargs_only
from scout_apm.core import objtrace
from scout_apm.core.agent.manager import CoreAgentManager
from scout_apm.core.agent.socket import CoreAgentSocketThread
from scout_apm.core.config import scout_config
from scout_apm.core.error_service import ErrorServiceThread
from scout_apm.core.metadata import report_app_metadata

logger = logging.getLogger(__name__)


@kwargs_only
def install(config=None):
    global shutdown_registered
    if config is not None:
        scout_config.set(**config)
    scout_config.log()

    if os.name == "nt":
        logger.info(
            "APM Not Launching on PID: %s - Windows is not supported", os.getpid()
        )
        return False

    if not scout_config.value("monitor"):
        logger.info(
            "APM Not Launching on PID: %s - Configuration 'monitor' is not true",
            os.getpid(),
        )
        return False

    instruments.ensure_all_installed()
    objtrace.enable()

    logger.debug("APM Launching on PID: %s", os.getpid())
    launched = CoreAgentManager().launch()

    report_app_metadata()
    if launched:
        # Stop the thread to avoid running threads pre-fork
        CoreAgentSocketThread.ensure_stopped()

    if scout_config.value("shutdown_timeout_seconds") > 0.0 and not shutdown_registered:
        atexit.register(shutdown)
        shutdown_registered = True

    return True


shutdown_registered = False


def shutdown():
    timeout_seconds = scout_config.value("shutdown_timeout_seconds")

    def apm_callback(queue_size):
        if scout_config.value("shutdown_message_enabled"):
            print(  # noqa: T001,T201
                (
                    "Scout draining {queue_size} event{s} for up to"
                    + " {timeout_seconds} seconds"
                ).format(
                    queue_size=queue_size,
                    s=("" if queue_size == 1 else "s"),
                    timeout_seconds=timeout_seconds,
                ),
                file=sys.stderr,
            )

    def error_callback(queue_size):
        if scout_config.value("shutdown_message_enabled"):
            print(  # noqa: T001,T201
                (
                    "Scout draining {queue_size} error{s} for up to"
                    + " {timeout_seconds} seconds"
                ).format(
                    queue_size=queue_size,
                    s=("" if queue_size == 1 else "s"),
                    timeout_seconds=timeout_seconds,
                ),
                file=sys.stderr,
            )

    CoreAgentSocketThread.wait_until_drained(
        timeout_seconds=timeout_seconds, callback=apm_callback
    )

    if scout_config.value("errors_enabled"):
        ErrorServiceThread.wait_until_drained(
            timeout_seconds=timeout_seconds, callback=error_callback
        )



# coding=utf-8

import wrapt
from starlette.background import BackgroundTask

import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import asgi_track_request_data


class ScoutMiddleware:
    def __init__(self, app):
        self.app = app
        installed = scout_apm.core.install()
        self._do_nothing = not installed
        if installed:
            install_background_instrumentation()

    async def __call__(self, scope, receive, send):
        if self._do_nothing or scope["type"] != "http":
            return await self.app(scope, receive, send)

        tracked_request = TrackedRequest.instance()
        # Assume the request is real until we determine it's not. This is useful
        # when the asyncio instrumentation is determining if a new Task should
        # reuse the existing tracked request.
        tracked_request.is_real_request = True
        # Can't name controller until post-routing - see final clause
        controller_span = tracked_request.start_span(operation="Controller/Unknown")

        asgi_track_request_data(scope, tracked_request)

        def grab_extra_data():
            if "endpoint" in scope:
                # Rename top span
                endpoint = scope["endpoint"]
                if not hasattr(endpoint, "__qualname__"):
                    endpoint = endpoint.__class__
                controller_span.operation = "Controller/{}.{}".format(
                    endpoint.__module__,
                    endpoint.__qualname__,
                )
                tracked_request.operation = controller_span.operation
            else:
                # Mark the request as not real
                tracked_request.is_real_request = False

            # From AuthenticationMiddleware - bypass request.user because it
            # throws AssertionError if 'user' is not in Scope, and we need a
            # try/except already
            try:
                username = scope["user"].display_name
            except (KeyError, AttributeError):
                pass
            else:
                tracked_request.tag("username", username)

        async def wrapped_send(data):
            type_ = data.get("type", None)
            if type_ == "http.response.start" and 500 <= data.get("status", 200) <= 599:
                tracked_request.tag("error", "true")
            elif type_ == "http.response.body" and not data.get("more_body", False):
                # Finish HTTP span when body finishes sending, not later (e.g.
                # after background tasks)
                grab_extra_data()
                tracked_request.stop_span()
            return await send(data)

        try:
            await self.app(scope, receive, wrapped_send)
        except Exception as exc:
            tracked_request.tag("error", "true")
            raise exc
        finally:
            if tracked_request.end_time is None:
                grab_extra_data()
                tracked_request.stop_span()


background_instrumentation_installed = False


def install_background_instrumentation():
    global background_instrumentation_installed
    if background_instrumentation_installed:
        return
    background_instrumentation_installed = True

    @wrapt.decorator
    async def wrapped_background_call(wrapped, instance, args, kwargs):
        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True

        operation = "Job/{}.{}".format(
            instance.func.__module__, instance.func.__qualname__
        )
        tracked_request.operation = operation
        with tracked_request.span(operation=operation):
            return await wrapped(*args, **kwargs)

    BackgroundTask.__call__ = wrapped_background_call(BackgroundTask.__call__)



# coding=utf-8

import os
import sys

from setuptools import Extension, find_packages, setup

with open("README.md", "r") as fp:
    long_description = fp.read()

packages = find_packages("src")

compile_extensions = (
    # Python 3+
    sys.version_info >= (3,)
    # Not Jython
    and not sys.platform.startswith("java")
    # Not PyPy
    and "__pypy__" not in sys.builtin_module_names
    # Not explicitly disabled
    and (os.environ.get("SCOUT_DISABLE_EXTENSIONS", "") == "")
)
if compile_extensions:
    ext_modules = [
        Extension(
            name=str("scout_apm.core._objtrace"),
            sources=[str("src/scout_apm/core/_objtrace.c")],
            optional=True,
        )
    ]
else:
    ext_modules = []

setup(
    name="scout_apm",
    version="3.5.3",
    description="Scout Application Performance Monitoring Agent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/scoutapp/scout_apm_python",
    project_urls={
        "Documentation": "https://docs.scoutapm.com/#python-agent",
        "Changelog": (
            "https://github.com/scoutapp/scout_apm_python/blob/master/CHANGELOG.md"
        ),
    },
    author="Scout",
    author_email="support@scoutapm.com",
    license="MIT",
    zip_safe=False,
    python_requires=">=3.8, <4",
    packages=packages,
    package_dir={str(""): str("src")},
    ext_modules=ext_modules,
    entry_points={
        "console_scripts": [
            "core-agent-manager = scout_apm.core.cli.core_agent_manager:main"
        ]
    },
    install_requires=[
        "asgiref",
        "psutil>=5",
        "urllib3",
        "certifi",
        "wrapt>=1.10,<2.0",
    ],
    keywords=["apm", "performance monitoring", "development"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Bottle",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Topic :: System :: Monitoring",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



# coding=utf-8

import datetime as dt
import logging
import re

from scout_apm.compat import iteritems

logger = logging.getLogger(__name__)

key_regex = re.compile(r"^[a-zA-Z0-9]{20}$")


def format_dt_for_core_agent(event_time: dt.datetime) -> str:
    """
    Returns expected format for Core Agent compatibility.
    Coerce any tz-aware datetime to UTC just in case.
    """
    # if we somehow got a naive datetime, convert it to UTC
    if event_time.tzinfo is None:
        logger.warning("Naive datetime passed to format_dt_for_core_agent")
        event_time = event_time.astimezone(dt.timezone.utc)
    return event_time.isoformat()


class Register(object):
    __slots__ = ("app", "key", "hostname")

    def __init__(self, app, key, hostname):
        self.app = app
        self.key = key
        self.hostname = hostname

    def message(self):
        key_prefix = self.key[:3]
        key_matches_regex = bool(key_regex.match(self.key))
        logger.info(
            "Registering with app=%s key_prefix=%s key_format_validated=%s host=%s"
            % (self.app, key_prefix, key_matches_regex, self.hostname)
        )
        return {
            "Register": {
                "app": self.app,
                "key": self.key,
                "host": self.hostname,
                "language": "python",
                "api_version": "1.0",
            }
        }


class StartSpan(object):
    __slots__ = ("timestamp", "request_id", "span_id", "parent", "operation")

    def __init__(self, timestamp, request_id, span_id, parent, operation):
        self.timestamp = timestamp
        self.request_id = request_id
        self.span_id = span_id
        self.parent = parent
        self.operation = operation

    def message(self):
        return {
            "StartSpan": {
                "timestamp": format_dt_for_core_agent(self.timestamp),
                "request_id": self.request_id,
                "span_id": self.span_id,
                "parent_id": self.parent,
                "operation": self.operation,
            }
        }


class StopSpan(object):
    __slots__ = ("timestamp", "request_id", "span_id")

    def __init__(self, timestamp, request_id, span_id):
        self.timestamp = timestamp
        self.request_id = request_id
        self.span_id = span_id

    def message(self):
        return {
            "StopSpan": {
                "timestamp": format_dt_for_core_agent(self.timestamp),
                "request_id": self.request_id,
                "span_id": self.span_id,
            }
        }


class StartRequest(object):
    __slots__ = ("timestamp", "request_id")

    def __init__(self, timestamp, request_id):
        self.timestamp = timestamp
        self.request_id = request_id

    def message(self):
        return {
            "StartRequest": {
                "timestamp": format_dt_for_core_agent(self.timestamp),
                "request_id": self.request_id,
            }
        }


class FinishRequest(object):
    __slots__ = ("timestamp", "request_id")

    def __init__(self, timestamp, request_id):
        self.timestamp = timestamp
        self.request_id = request_id

    def message(self):
        return {
            "FinishRequest": {
                "timestamp": format_dt_for_core_agent(self.timestamp),
                "request_id": self.request_id,
            }
        }


class TagSpan(object):
    __slots__ = ("timestamp", "request_id", "span_id", "tag", "value")

    def __init__(self, timestamp, request_id, span_id, tag, value):
        self.timestamp = timestamp
        self.request_id = request_id
        self.span_id = span_id
        self.tag = tag
        self.value = value

    def message(self):
        return {
            "TagSpan": {
                "timestamp": format_dt_for_core_agent(self.timestamp),
                "request_id": self.request_id,
                "span_id": self.span_id,
                "tag": self.tag,
                "value": self.value,
            }
        }


class TagRequest(object):
    __slots__ = ("timestamp", "request_id", "tag", "value")

    def __init__(self, timestamp, request_id, tag, value):
        self.timestamp = timestamp
        self.request_id = request_id
        self.tag = tag
        self.value = value

    def message(self):
        return {
            "TagRequest": {
                "timestamp": format_dt_for_core_agent(self.timestamp),
                "request_id": self.request_id,
                "tag": self.tag,
                "value": self.value,
            }
        }


class ApplicationEvent(object):
    __slots__ = ("event_type", "event_value", "source", "timestamp")

    def __init__(self, event_type, event_value, source, timestamp):
        self.event_type = event_type
        self.event_value = event_value
        self.source = source
        self.timestamp = timestamp

    def message(self):
        return {
            "ApplicationEvent": {
                "timestamp": format_dt_for_core_agent(self.timestamp),
                "event_type": self.event_type,
                "event_value": self.event_value,
                "source": self.source,
            }
        }


class BatchCommand(object):
    __slots__ = ("commands",)

    def __init__(self, commands):
        self.commands = commands

    def message(self):
        return {
            "BatchCommand": {
                "commands": [command.message() for command in self.commands]
            }
        }

    @classmethod
    def from_tracked_request(cls, request):
        # The TrackedRequest must be finished
        commands = []
        commands.append(
            StartRequest(timestamp=request.start_time, request_id=request.request_id)
        )
        for key, value in iteritems(request.tags):
            commands.append(
                TagRequest(
                    timestamp=request.start_time,
                    request_id=request.request_id,
                    tag=key,
                    value=value,
                )
            )

        for span in request.complete_spans:
            commands.append(
                StartSpan(
                    timestamp=span.start_time,
                    request_id=span.request_id,
                    span_id=span.span_id,
                    parent=span.parent,
                    operation=span.operation,
                )
            )

            for key, value in iteritems(span.tags):
                commands.append(
                    TagSpan(
                        timestamp=span.start_time,
                        request_id=request.request_id,
                        span_id=span.span_id,
                        tag=key,
                        value=value,
                    )
                )

            commands.append(
                StopSpan(
                    timestamp=span.end_time,
                    request_id=span.request_id,
                    span_id=span.span_id,
                )
            )

        commands.append(
            FinishRequest(timestamp=request.end_time, request_id=request.request_id)
        )

        return cls(commands)
